/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef DECRYPTHD_DECRYPT_EVOB_H
#define DECRYPTHD_DECRYPT_EVOB_H

#include "logger.hpp"
#include <string>
#include <fstream>
#include <stdint.h>

// AES CBC init vector
static const uint8_t AESCBCConstantIV[] = {
	0x0B, 0xA0, 0xF8, 0xDD, 0xFE, 0xA6, 0x1F, 0xB3, 0xD8, 0xDF, 0x9F, 0x56, 0x6A, 0x05, 0x0F, 0x78
};

/**
 * Process entire EVOB file.
 *
 * File is split into chunks and each chunk is chopped into EVOB packs of 2048
 * bytes. These are then ecrypted if needed and written to a new file.
 *
 * Contains some compiler optimizations, such as quick memcopies and loop
 * unrolling. Might look a bit weird, but for a 10GB file this code gets
 * executed about 5 million times ;)
 */
class EVOBFile
{
public:
	EVOBFile(Logger* m_pLog, bool writeOutput, bool dumpOutput, unsigned long chunkSize);

	void 		Decrypt		(const std::string&		sourceFile,
							 const std::string& 	destinationDir,
							 const uint8_t* const	pTitleKeyTable);
	void		Open		(const std::string& 	sourceFile,
							 const uint8_t* const 	pTitleKeyTable,
							 std::ifstream&			EVOBStream);
	int			ReadChunk	(std::ifstream&			EVOBStream,
							 uint8_t* const			currentChunk);

private:
	EVOBFile(Logger* m_pLog);

	int 		LoadChunk	(uint8_t* const			pBuffer,
							 std::ifstream& 		inputStream,
							 int 					blockLen);
	void 		DecryptPack(uint8_t* const			rawData);

	const static unsigned int EVOBSize			= 2048;
	const static unsigned int EVOBHeaderSize	= 128;
	const static unsigned int EVOBContentSize	= 1920;

/*
#define PACK_START_CODE             ((unsigned int)0x000001ba)
#define SYSTEM_HEADER_START_CODE    ((unsigned int)0x000001bb)
#define SEQUENCE_END_CODE           ((unsigned int)0x000001b7)
#define PACKET_START_CODE_MASK      ((unsigned int)0xffffff00)
#define PACKET_START_CODE_PREFIX    ((unsigned int)0x00000100)
#define ISO_11172_END_CODE          ((unsigned int)0x000001b9)

// mpeg2
#define PROGRAM_STREAM_MAP 0x1bc
#define PRIVATE_STREAM_1   0x1bd
#define PADDING_STREAM     0x1be
#define PRIVATE_STREAM_2   0x1bf
*/
	enum ePackHeader
	{
		PACK_START_CODE				= 0x000001BA,	//
		SYSTEM_HEADER_START_CODE	= 0x000001BB,	//
		PROGRAM_STREAM_MAP			= 0x000001BC,	// MPEG2
		PRIVATE_STREAM_1			= 0x000001BD,	// MPEG2
		PADDING_STREAM				= 0x000001BE,	// MPEG2
		PRIVATE_STREAM_2			= 0x000001BF	// MPEG2
	};

	enum eSubstreamID
	{
		PCI_PACKET			= 0x00,			// PCI packet
		DSI_PACKET			= 0x01,			// DSI packet
		GCI_PACKET			= 0x04,			// GCI packet
		HLI_PACKET			= 0x08,			// HLI packet
		ADV_PACKET			= 0x80			// ADV packet
	};

	Logger*			m_pLog;					/**< pointer to logger */
	bool			m_writeOutput;			/**< true writes output to disk */
	bool			m_dumpOutput;			/**< true to dump output to stdout */
	unsigned long 	m_chunkSize;			/**< chcunksize (32*1048576; 33.554.432, testfile 32.429.056) */
	uint8_t 		m_EVOBHeader[EVOBHeaderSize];	/**<  */
	uint8_t 		m_EVOBContent[EVOBContentSize];	/**<  */
	int 			m_TITLE_KEY_PTR;		/**< TITLE_KEY_PTR */
	uint8_t 		m_CPIField[16];			/**< CPI field */
	uint8_t 		m_TitleKeyData[16];		/**< Title Key data */
	uint8_t 		m_TitleKey[16];			/**< single Title Key */
	uint8_t*		m_pTitleKeyTable;		/**< Title Key table */
};

#endif
