/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "acadecrypter.hpp"
#include "aacs.hpp"
#include "volumekeyloader.hpp"
#include "acafile.hpp"
#include "util.hpp"
#include <sstream>
#include <stdexcept>

/**
 *
 */
ACADecrypter::ACADecrypter(Logger& oLog)
:	m_oLog(oLog)
{
}

/**
 * Decrypt ACA file
 */
void ACADecrypter::DecryptFile(const std::string& filename,
							   const std::string& destinationDir)
{
	/** @todo Check for trailing slash */
	const std::string fullFilePath			= m_mountPoint + "ADV_OBJ/" + filename;
	const std::string fullDestinationPath	= destinationDir;
	const std::string TitleKeyFilename 		= m_mountPoint + "AACS/VTKF000.AACS";

	try
	{
		// fire up AACS engine
		m_oLog.LogInfo("Initialising AACS engine");
		std::auto_ptr<AACS> AACSEngine( new AACS(&m_oLog, false, m_device, m_mountPoint));
		AACSEngine->DetectMedia();

		// load Title Key file, get VUK and decrypt titlekeys
		m_oLog.LogInfo("Loading TitleKey file and getting Volume Key from keydb");
		AACSEngine->LoadTitleKeyFile();

		VolumeKeyLoader KeyLoader(&m_oLog);
		const std::string strVolumeKey = KeyLoader.GetVolumeKey(TitleKeyFilename, m_keyDBFilename);

		if (strVolumeKey.empty()) {
			throw std::runtime_error("No Volume Key found...");
		}
		m_oLog.LogInfo("Volume Key found:\t\t" + strVolumeKey);

		AACSEngine->DecryptTitleKeys(strVolumeKey);

		// start decrypting
		m_oLog.LogInfo("Starting actual decryption of " + filename);
		m_oLog.LogInfo("Destination: " + fullDestinationPath);
		ACAFile encryptedFile(&m_oLog);
		encryptedFile.Decrypt(fullFilePath, fullDestinationPath, AACSEngine->TitleKeyTable(), m_writeToDisk);
		m_oLog.LogInfo("Decryption of " + filename + " completed succesfully");

		std::stringstream message;
		message << encryptedFile.DecryptedFilesize();
		m_oLog.LogDebug(message.str());
	}
	catch (...)
	{
		throw;	// just pass on
	}
}

//----------------------

void ACADecrypter::SetKeyDB(const std::string& keyDBFilename)
{
	m_keyDBFilename = keyDBFilename;
}
void ACADecrypter::SetDevice(const std::string& device)
{
	m_device = device;
}
void ACADecrypter::SetMountPoint(const std::string& mountPoint)
{
	m_mountPoint = mountPoint;
	Utils::AddSlash(m_mountPoint);
}
void ACADecrypter::SetWriteOutput(bool writeToDisk)
{
	m_writeToDisk = writeToDisk;
}
