/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string>
#include <iostream>
#include <fstream>

#include "volumekeyloader.hpp"
#include "aacs.hpp"
#include "util.hpp"
#include "sha1.hpp"
#include "Markup.h"

VolumeKeyLoader::VolumeKeyLoader(Logger* pLog)
:	m_pLog(pLog)
{
}

/**
 * Open XML keydb and get Volume Key
 */
const std::string VolumeKeyLoader::GetVolumeKey(const std::string& TKFile, const std::string& keyDb)
{
	// get TKFMAC
	std::string TKFMAC = GetTKFMAC(TKFile);
	m_pLog->LogInfo("Title Key File MAC:\t" + TKFMAC);

	// get TKF SHA-1 hash - store in TKFMAC
	std::string SHA1 = GetTKFHash(TKFile);
	m_pLog->LogInfo("Title Key File hash:\t" + SHA1);

	// detect filetype by extention
	//std::string fileType;
	//todo

	if (keyDb == "keydb.xml") {
		return LoadXml(keyDb, SHA1, TKFMAC);
	}
	else if (keyDb == "keydb.cfg") {
		return LoadCfg(keyDb, SHA1);
	}

	return "";
}

/**
 * Do SHA1 magic on file
 */
const std::string VolumeKeyLoader::GetTKFHash(const std::string& TKFile) const
{
	//Log->Msg("Getting TKFHash from " + TKFile);

	std::ifstream fis(TKFile.c_str());
	if (!fis.is_open()) {
		m_pLog->LogError(TKFile + " not found");
		return 0;
	}
	fis.close();

	unsigned char * hashBuffer = new unsigned char[SHA1_LENGTH];
	CSHA1 sha1;

	try
	{
		sha1.Reset();
		sha1.HashFile((char*)TKFile.c_str());
		sha1.Final();
		sha1.GetHash(hashBuffer);
	}
	catch (std::exception e) {
		throw;
	}

	//char hash[SHA1_LENGTH*2];
	std::string strHash;
	Utils::ToHexString(hashBuffer, strHash, SHA1_LENGTH);

	delete[] hashBuffer;
	//return hash;
	return strHash;
}

/**
 * Get TKF MAC
 */
const std::string VolumeKeyLoader::GetTKFMAC(const std::string& strTKFile) const
{
	//Log->Msg("Getting TKFMAC from " + TKFile);

	unsigned char *	TKFMAC = new unsigned char[MAC_LENGTH];
	//char *			MAC    = new char[32];

	std::ifstream fis(strTKFile.c_str(), std::ios::binary | std::ios::in);
	if (!fis.is_open()) {
		m_pLog->LogError("Error accessing " + strTKFile + " while trying to get TKF MAC!");
		return 0;
	}

	std::streambuf *pBuf = fis.rdbuf();
	pBuf->pubseekpos(2464);
	pBuf->sgetn((char*)TKFMAC, MAC_LENGTH);
	std::string strMAC;
	Utils::ToHexString(TKFMAC, strMAC, MAC_LENGTH);

	delete[] TKFMAC;
	return strMAC;
}


/**
 * search for Volume Key in XML keydb
 */
const std::string VolumeKeyLoader::LoadXml(const std::string& fileName, const std::string& SHA1, const std::string& TKFMAC)
{
	CMarkup xml;

	if (!xml.Load( fileName.c_str() )) {
		m_pLog->LogError("Error loading " + fileName);
		return "";
	}

	xml.ResetPos();
	bool macFound  = false;
	bool hashFound = false;

	// loop through all field sections
	while (xml.FindChildElem("KeySet"))
	{
		xml.IntoElem();

		while (xml.FindChildElem("Key"))
		{
			if (xml.GetChildAttrib("type") == "CMAC") {
				if (xml.GetChildData() == TKFMAC) macFound = true;
			}

			if (xml.GetChildAttrib("type") == "Hash" && (macFound == false)) {
				if (xml.GetChildData() == SHA1) hashFound = true;
			}

			if (xml.GetChildAttrib("type") == "Volume" && (macFound || hashFound)) {
				return xml.GetChildData();
			}
		}

		//if (xml.GetChildAttrib("type") == "Description" && (macFound || hashFound)) {
		//	cout << "Descr=" << xml.GetChildData() << endl;
		//}

		// exit current KeySet section
		xml.OutOfElem();
	}
}

/**
 *
 */
const std::string VolumeKeyLoader::LoadCfg(const std::string& fileName, const std::string& SHA1)
{
	std::string		movie;
	std::string		hash;
	std::string		type;
	std::string		timestamp;
	std::string		VolumeKey;
	std::ifstream	keyDb(fileName.c_str());

	if (!keyDb.is_open()) {
		m_pLog->LogError("Error loading " + fileName);
		return "";
	}

	while(!std::getline(keyDb, movie, '|').eof())
	{
		Utils::Trim(movie);
		hash = movie.substr(0, 40);
		std::cout << movie << ": " << hash << std::endl;

		if (hash == SHA1)
		{
			std::getline(keyDb, type, '|');
			std::getline(keyDb, timestamp, '|');
			std::getline(keyDb, VolumeKey);
			return VolumeKey;
		}
	}

	return "";
}
