/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * Decrypt HD by Zotty
 * A program for decrypting AACS realtime
 *
 * This version is intended to use files stored on harddisk. It uses a
 * keydatabase to get the keys.
 */

#include "evobdecrypter.hpp"
#include "logger.hpp"
#include <iostream>
#include <sstream>
#include <stdexcept>

void Usage();

/**
 *
 */
int main (int argc, const char* argv[])
{
	if (argc < 5)
	{
		Usage();
		return 0;
	}

	std::stringstream appString;
	appString << version << " (" << __DATE__ << " " << __TIME__ << ")";
	bool			writeOutput		= true;
	bool			dumpOutput		= false;
	unsigned long	chunkSize		= 8*2048;		// 64KB (=32*2048), 32MB = 32*1048576
	std::string 	destinationDir;					// output is dumped here

	// process args
	std::string keyDBFilename	= argv[1];			// keydb.xml
	std::string mountPoint		= argv[2];			// directory containing HDDVD files
	std::string filename		= argv[3];			// file to decrypt

	// check if we want streaming
	std::string arg = argv[4];
	if (arg == "-o")
	{
		dumpOutput 	= true;			// set flag to dump to stdout
		writeOutput	= false;		// no writing to file
	}
	else
	{
		destinationDir = arg;		// write to file
		// check again for -o after this for filewriting while watching
	}

	// additional options
	// argv starts at 0 while argc starts at 1... sigh
	for (unsigned short i = 5; i < argc; i++)
	{
		arg = argv[i];

		// watch and write
		if (arg == "-o") {
			dumpOutput = true;			// set flag to dump to stdout
		}

		// chunksize
		if (arg == "-c")
		{
			if (i + 1 <= argc)
			{
				i++;
				chunkSize = strtol(argv[i], 0, 10);
			}
			else
				Usage();
		}
	}

	Logger oLog("decrypthd_nodrive.log");
	oLog.SetLevel(Logger::DEBUG);
	oLog.LogInfo("DecryptHD "		+ appString.str());
	oLog.LogInfo("KeyDB:\t\t"		+ keyDBFilename);
	oLog.LogInfo("Mountpoint:\t"	+ mountPoint);
	oLog.LogInfo("Filename:\t\t"	+ filename);

	if (dumpOutput) {
		oLog.LogInfo("Ouput dir: \tdumping to stdout");
	}

	if (writeOutput) {
		oLog.LogInfo("Ouput dir:\t" + destinationDir);
	}

	std::stringstream msg;
	msg << "Chunksize:\t\t" << chunkSize << std::endl;
	oLog.LogInfo(msg.str());

	try
	{
		EVOBDecrypter Decrypter(oLog);
		//Decrypter.SetDevice(strDevice);
		Decrypter.SetMountPoint(mountPoint);
		Decrypter.SetDestinationPath(destinationDir);
		Decrypter.SetFileName(filename);
		Decrypter.SetKeyDB(keyDBFilename);
		Decrypter.SetChunkSize(chunkSize);
		Decrypter.SetWriteOutput(writeOutput);
		Decrypter.SetDumpOutput(dumpOutput);
		Decrypter.StartDecrypting(false);
		oLog.LogInfo("Finished decrypting");
	}
	catch (const std::exception &e)
	{
		std::stringstream errorMsg;
		errorMsg << "\n### Exception: " << e.what() << " ###";
		oLog.LogError(errorMsg.str());
	}
	catch (...)
	{
		oLog.LogError("\n### Unspecified exception occured ###");
	}

	return 0;
}

/**
 * Show usage msg
 */
void Usage()
{
	std::cout 	<<	"Usage: decrypthd <keydb> <mountpoint> <filename> -o [options] | <player>\n"
					"\n"
					"Options:\n"
				 	"-c\tChunksize"
				 	"\n"
				 	"Example: decrypthd keydb.xml ./disc/ file.evo -o | mplayer -demuxer lavf -vc -ffvc1 -\n";
}
