/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * EVOB decrypter by Zotty
 * A program for decrypting AACS realtime
 */

#include "evobdecrypter.hpp"
#include "volumekeyloader.hpp"
#include "aacs.hpp"
#include "evobfile.hpp"
#include "filesystem.hpp"
#include "util.hpp"
#include <iostream>
#include <sstream>
#include <stdexcept>

/**
 *
 */
EVOBDecrypter::EVOBDecrypter(Logger& oLog)
:	m_oLog(oLog),
	m_writeOutput(true),
	m_dumpOutput(false)
{
}

/**
 * Start HD decryption
 */
bool EVOBDecrypter::StartDecrypting(bool useDriveAuthentication)
{
	m_oLog.LogInfo("*** Setting up decrypter ***");
	const std::string TitleKeyFilename = m_mountPoint + "AACS/VTKF000.AACS";

	try
	{
		std::auto_ptr<AACS> AACSEngine( new AACS(&m_oLog, useDriveAuthentication, m_device, m_mountPoint) );
		AACSEngine->DetectMedia();

		// load Title Key file, get VUK and decrypt titlekeys
		m_oLog.LogInfo("Loading TitleKey file and getting Volume Key from keydb");
		AACSEngine->LoadTitleKeyFile();

		VolumeKeyLoader KeyLoader(&m_oLog);
		const std::string strVolumeKey = KeyLoader.GetVolumeKey(TitleKeyFilename, m_keyDBFilename);

		if (strVolumeKey.empty()) {
			throw std::runtime_error("No Volume Key found...");
		}
		m_oLog.LogInfo("Volume Key found:\t\t" + strVolumeKey);

		AACSEngine->DecryptTitleKeys(strVolumeKey);

		// decrypt file
		DecryptFile(m_filename, AACSEngine.get());
	}
	catch (const std::exception &e)
	{
		std::stringstream errorMsg;
		errorMsg << "EVOB ecrypter: " << e.what();
		throw std::runtime_error(errorMsg.str());
	}
	catch (...) {
		throw std::runtime_error("### EVOB decrypter: Unspecified exception occured ###");
	}

	return true;
}

/**
 * Decrypt single file
 */
void EVOBDecrypter::DecryptFile(std::string filename, AACS* AACSEngine)
{
	m_oLog.LogInfo("*** Starting decryption of " + filename + " ***");
	const std::string fullFilePath			= m_mountPoint + "HVDVD_TS/" + filename;
	const std::string fullDestinationPath 	= m_destinationPath + filename;

	if (m_dumpOutput)	{ m_oLog.LogInfo("Dumping to stdout"); }
	else				{ m_oLog.LogInfo("Saving to " + fullDestinationPath); }

	// get keynr
	const int keyNumber = AACSEngine->GetKeyFromEVOB(fullFilePath);
	std::stringstream msg;
	msg << "Needed key is " << keyNumber;
	m_oLog.LogDebug(msg.str());

	// proceed only if we have a keynr
	if (keyNumber < 0) {
		throw std::runtime_error("No keynumber. Aborting!");
	}

	// check if we have the (decrypted) Title Key
	if (AACSEngine->HasKey(keyNumber)) {
		EVOBFile encryptedFile(&m_oLog, m_writeOutput, m_dumpOutput, m_chunkSize);
		encryptedFile.Decrypt(fullFilePath, fullDestinationPath, AACSEngine->TitleKeyTable());
	}
	else {
		throw std::runtime_error("Could not retrieve Title Key. Aborting!");
	}
}

//------------------------- ACCESSORS ---------------------------

void EVOBDecrypter::SetDevice(const std::string& device)
{
	m_device = device;
}
void EVOBDecrypter::SetMountPoint(const std::string& mountPoint)
{
	m_mountPoint = mountPoint;
	Utils::AddSlash(m_mountPoint);
}
void EVOBDecrypter::SetDestinationPath(const std::string& destinationPath)
{
	m_destinationPath = destinationPath;
	Utils::AddSlash(m_destinationPath);
}
void EVOBDecrypter::SetFileName(const std::string& filename)
{
	m_filename = filename;
}
void EVOBDecrypter::SetKeyDB(const std::string& keyDBFilename)
{
	m_keyDBFilename = keyDBFilename;
}
void EVOBDecrypter::SetWriteOutput(bool writeOutput)
{
	m_writeOutput = writeOutput;
}
void EVOBDecrypter::SetDumpOutput(bool dumpOutput)
{
	m_dumpOutput = dumpOutput;
}
void EVOBDecrypter::SetChunkSize(unsigned long chunkSize)
{
	m_chunkSize = chunkSize;
}
