#include "hddvdfs.hpp"
#include "evobfile.hpp"
#include "acafile.hpp"
#include "arfile.hpp"
#include "logger.hpp"
#include <iostream>
#include <sstream>
#include <csignal>
#include <stdexcept>
#include <sys/types.h>

#define CHUNK 			65536
#define FIX_FILE_MODES	1		// Make all files and directores read+exec
#define MOUNT_TIMEOUT	10
#define MIN(x,y) ((x) < (y) ? (x) : (y))

HdDvdFs *alarm_obj;
Logger*	pLog = new Logger("/tmp/hddvdfs.log");
std::map<std::string, struct stat> HdDvdFs::fileMap;

using namespace std;

static void alarm_handler(int sig)
{
	alarm_obj->unmount();
}

static int my_system(const string cmd) {
	std::cerr << "Running command: " << cmd << std::endl;
	return system(cmd.c_str());
}

int HdDvdFs::try_unmount() {
	return my_system("umount "+self->m_UdfMountPoint);
}

/**
 *
 */
void HdDvdFs::unmount()
{
	if (self->m_mounted)
	{
		::closedir(self->m_dir);
		if (self->try_unmount() == 0) {
			delete self->m_disc;
			self->m_mounted = false;
		}
		else {
			self->m_dir = ::opendir(self->m_UdfMountPoint.c_str());
			alarm(MOUNT_TIMEOUT); // Try again ...
		}
	}
}

// Constructor
HdDvdFs::HdDvdFs (string dev)
{
    m_Device 	= dev;
    m_refcnt 	= 0;
    m_mounted	= false;
    signal(SIGALRM, alarm_handler);
    alarm_obj 	= this;

	pLog->SetLevel(Logger::DEBUG);
}

int HdDvdFs::addref()
{
    if (self->m_refcnt < 1)
	{
		alarm(0);

		if (!m_mounted)
		{
			if (my_system("mount -r "+self->m_UdfMountPoint) != 0 &&
				my_system("mount -r "+self->m_Device+" "+self->m_UdfMountPoint) != 0)
			{
				return -ENODEV;
			}

			std::stringstream ssTitleKeyFile;
			self->m_disc = new AACS(pLog, true, self->m_Device, self->m_UdfMountPoint + "/");
			self->m_disc->DetectMedia();
			self->m_disc->LoadTitleKeyFile();
			self->m_disc->DecryptTitleKeys("");
			self->m_dir = ::opendir(self->m_UdfMountPoint.c_str());
			m_mounted = true;
		}
		self->m_refcnt = 1;
    }
    else
		self->m_refcnt++;
    return 0;
}

void HdDvdFs::delref()
{
	if (self->m_refcnt < 1) {
		std::cerr << "warning: reference count underflow" << std::endl;
		return;
    }

    self->m_refcnt--;
    if (self->m_refcnt == 0) {
		if (m_mounted)
	    	alarm(MOUNT_TIMEOUT);
    }
}

void HdDvdFs::set_mountpoint (string path)
{
	// Trim slashes from the end of path
	while (path.rfind('/') == path.length()-1)
	    path.resize(path.length()-1);

	self->m_MountPoint = path;

	// Make the udf mount point have a "." in front of the final dirname.
	int i = path.rfind('/');
	if (i == string::npos)
	    i = -1;
	self->m_UdfMountPoint = path.substr(0,i+1)+"."+path.substr(i+1);

	std::cerr << "mkdir: " << self->m_UdfMountPoint << std::endl;
	::mkdir(self->m_UdfMountPoint.c_str(), 0777);

	// Try this in case we have a leftover mount on this dir.
	(void) self->try_unmount();
}

class HdDvdFsFH
{
public:
	HdDvdFsFH()
    : 	evob(0),
		aca(0),
		arf(0),
		chunk_ptr(0),
		chunk_offset(0),
		chunk_end(0)
    {
	}

	~HdDvdFsFH() {
	    if (evob != 0)
 			delete evob;
		if (aca != 0)
			delete aca;
		if (arf != 0)
			delete arf;
	    if (chunk_ptr != 0)
			delete chunk_ptr;
	}

	EVOBFile		*evob;
	ACAFile			*aca;
	ARFile			*arf;
	std::ifstream	fis;
	uint8_t			*chunk_ptr;		// pointer to chunk buffer
	off_t			chunk_offset;	// offset in chunk
	size_t			chunk_end;		// end of valid data in chunk
};

/**
 * Get file attributes
 *
 * This function is called extensively during regular filesystem operation which
 * means that it is crucial that is is implemented efficiently. This means that
 * you should, in most cases, build a attribute cache such as a hashtable using
 * the full path as key and the struct stat as value.
 *
 * @param path	File path
 * @param stbuf	stat struct were file info is stored
 */
int HdDvdFs::getattr (const char *path, struct stat *stbuf)
{
	string strPath(path);
	string fullPath = self->m_UdfMountPoint + "/" + strPath;

	if (self->addref() != 0)
	    return -ENODEV;
	self->delref();

	// fill attributes
	if (::stat(fullPath.c_str(), stbuf) < 0)
	    return -errno;
#ifdef FIX_FILE_MODES
	stbuf->st_mode |= 0555;
#endif

	// take care of files that get stripped of AACS header
	if (fileMap.empty())
		return 0;

	// overwrite size if file is listed
	std::map<std::string, struct stat>::const_iterator iter;
	iter = fileMap.find(strPath);
	if (iter != fileMap.end()) {
		std::cerr << "Found file in map: original size = " << stbuf->st_size
			<< ", updated size = " << iter->second.st_size << std::endl;
		stbuf->st_size = iter->second.st_size;
	}

	return 0;
}

/**
 * Read directory contents
 *
 * @param path		Path to directory
 * @param buf		Buffer containing information about directory
 * @param filler	Pointer to fuse_fill_dir_t() function that is used to add entries to buffer
 * @param offset	Offset in directory entries
 * @param fi		Pointer to fuse_file_info
 */
int HdDvdFs::readdir(const char *path, void *buf, fuse_fill_dir_t filler,
					off_t offset, struct fuse_file_info *fi)
{
	string strPath (path);
	string fullPath = self->m_UdfMountPoint + "/" + strPath;

	if (self->addref() != 0)
	    return -ENODEV;
	self->delref();

	DIR *dir = ::opendir(fullPath.c_str());
	if (!dir)
	    return -errno;

        filler(buf, ".", NULL, 0);
        filler(buf, "..", NULL, 0);

	struct dirent *ent;
	while (ent = ::readdir(dir))
	    filler(buf, ent->d_name, NULL, 0);

	closedir(dir);

	return 0;
}

/**
 * Open file
 *
 * This function is called to check if user is permitted to open the file using
 * given flags.
 *
 * @param path	File path
 * @param fi	Pointer to fuse_file_info
 */
int HdDvdFs::open(const char *path, struct fuse_file_info *fi)
{
	string strPath (path);
	string fullPath = self->m_UdfMountPoint + "/" + strPath;
	int retval = 0;

	//
	if (self->addref() != 0)
	    return -ENODEV;

	HdDvdFsFH *fh = new HdDvdFsFH();

	// convert extention to lowercase so we know for sure all extenstions get caught
	std::string extention = strPath.substr( strPath.rfind('.', strPath.length()) );
	transform(extention.begin(), extention.end(), extention.begin(), (int(*)(int))tolower);

	// EVO files
	if (strPath.find("/HVDVD_TS") == 0 && extention == ".evo")
	{
	    std::cerr << "Using AACS decryption for " << strPath << std::endl;

		// get needed key number
	    const int keyNumber = self->m_disc->GetKeyFromEVOB(fullPath);
	    if (keyNumber < 0) {
		    std::cerr << "No key for this file" << std::endl;
		    retval = -EACCES;
	    }
	    else if (!self->m_disc->HasKey(keyNumber)) {
		    std::cerr << "Could not retrieve Title Key." << std::endl;
		    retval = -EACCES;
	    }
	    else
		{
			// open EVOB
			fh->evob = new EVOBFile(pLog, false, false, CHUNK);
			fh->evob->Open(fullPath, self->m_disc->TitleKeyTable(), fh->fis);

			// alloc memory
			fh->chunk_ptr = new uint8_t[CHUNK];
	    }
	}

	// ACA files
	else if (strPath.find("/ADV_OBJ") == 0 && extention == ".aca")
	{
	    std::cerr << "Stripping AACS from " << strPath << std::endl;
		fh->aca = new ACAFile(pLog);
		fh->aca->Open(fullPath, self->m_disc->TitleKeyTable(), fh->fis);

		// update filesize
		struct stat stbuf;
		stbuf.st_size = fh->aca->DecryptedFilesize();
		fileMap.insert( std::make_pair(strPath, stbuf) );
	}

	// XPL files (not encrypted, but have AACS header)
	else if (strPath.find("/ADV_OBJ") == 0 && extention == ".xpl")
	{
		std::cerr << "Stripping AACS header from " << strPath << std::endl;

		// create ARF
		// by creating it, it'll be read and stripped. The final result stays in
		// memory until deleted
		fh->arf = new ARFile(pLog, fullPath);

		// update filesize
		struct stat stbuf;
		stbuf.st_size = fh->arf->Filesize();
		fileMap.insert( std::make_pair(strPath, stbuf) );
	}

	// the rest
	else {
		std::cerr << "No decryption for " << strPath << std::endl;
	    fh->fis.open(fullPath.c_str(), std::ios::binary);
	}

	//
	if (retval == 0)
	    fi->fh = (uint64_t)fh;
	else {
	    delete fh;
	    self->delref();
	}

	// 0 on success
	return retval;
}

/**
 * Release file
 */
int HdDvdFs::release(const char *path, struct fuse_file_info *fi)
{
	if (fi->fh)
	{
		HdDvdFsFH *fh = (HdDvdFsFH *)fi->fh;
		delete fh;
		fi->fh = 0;
		self->delref();
	}

	return 0;
}

/**
 * Read data from already opened file
 *
 * Read information beginning in offset of size length and put it to buf.
 *
 * Decrypted files with a new filesize do not always have the new size since
 * read() does not call getattr() first!
 *
 * @param path		File path
 * @param buf		Buffer to read data into
 * @param size		No of bytes to read
 * @param offset	Offset in file
 * @param fi		Pointer to fuse_file_info
 * @return No of bytes read, or negeted error number on error
 */
int HdDvdFs::read(const char *path, char *buf, size_t size, off_t offset,
                struct fuse_file_info *fi)
{
	HdDvdFsFH*	fh = (HdDvdFsFH *)fi->fh;		// handle to fileinfo
	size_t		bytes_read = 0;					// no of bytes read

	// as long as there is more data to read
	while (bytes_read < size)
	{
		size_t 	n;								//
		size_t	need = size   - bytes_read;		// no of bytes remaining to read
		off_t	pos  = offset + bytes_read;		// current offset in file

		// EVOB files
		if (fh->evob)
		{
			// read less than entire chunk
			if (need < CHUNK)
			{
				// ?
				if (pos < fh->chunk_offset || pos >= fh->chunk_end)
				{
					fh->fis.seekg(pos);						// set offset in original file
					n = fh->evob->ReadChunk(fh->fis, fh->chunk_ptr);	// read decrypted data
					if (n < 0)								// we actually read data?
						return -errno;
					fh->chunk_offset = pos;					// update offset
					fh->chunk_end 	 = pos + n;				// update end of valid data
				}

				size_t chunk_left = fh->chunk_end - pos;	//
				n = MIN(need, chunk_left);					//

				//
				memcpy(buf + bytes_read, fh->chunk_ptr + (pos - fh->chunk_offset), n);
			}

			// read entire chunk
			else {
				fh->fis.seekg(pos);
				n = fh->evob->ReadChunk(fh->fis, (uint8_t*)buf + bytes_read);
			}
	    }

		// aca files
		else if (fh->aca)
		{
			// data is already decrypted, so just read into buffer

			// check size
			if (pos + need > fh->aca->DecryptedFilesize())
			{
				size_t bytes_remaining = fh->aca->DecryptedFilesize() - pos;
				std::cerr << "bytes_remaining=" << bytes_remaining << std::endl;
				n = fh->aca->ReadChunk(pos, bytes_remaining, (uint8_t*)buf + bytes_read);
			}
			else
			{
				std::cerr << "need=" << need << std::endl;
				n = fh->aca->ReadChunk(pos, need, (uint8_t*)buf + bytes_read);
			}
		}

		// ARF files
		else if (fh->arf)
		{
			// data is already decrypted, so just read into buffer

			// check size
			if (pos + need > fh->arf->Filesize())
			{
				size_t bytes_remaining = fh->arf->Filesize() - pos;
				//std::cerr << "bytes_remaining=" << bytes_remaining << std::endl;
				n = fh->arf->ReadChunk(pos, bytes_remaining, (uint8_t*)buf + bytes_read);
			}
			else
			{
				//std::cerr << "need=" << need << std::endl;
				n = fh->arf->ReadChunk(pos, need, (uint8_t*)buf + bytes_read);
			}
		}

		// non AACS files
		else {
			fh->fis.seekg(pos);
			fh->fis.read(buf+bytes_read, need);
			n = fh->fis.gcount();
	    }

	    if (n < 1)
		{
			if (n == 0)
				break;
			return -errno;
	    }

		// update no of bytes read
	    bytes_read += n;
	}

	return bytes_read;
}

/**
 *
 */
int main (int argc, char **argv)
{
	struct fuse *fuse;
 	char *mountpoint;
	int multithreaded;
	int res;

	HdDvdFs fs(argv[1]);

	// The first 3 parameters are identical to the fuse_main function.
	// The last parameter gives a pointer to a class instance, which is
	// required for static methods to access instance variables/ methods.

	// We've disabled the fuse_main call in fusexx, so this call
	// only does object initialization.

	(void) fs.main (0, NULL, NULL, &fs);

	// The following code is based on fuse_main_common in fuse lib.

	argc--;
	argv[1] = argv[0];

	fuse = fuse_setup(argc, argv+1, &HdDvdFs::operations, sizeof(HdDvdFs::operations), &mountpoint, &multithreaded, NULL);
	if (fuse == NULL)
	    return 1;

	// hddvdfs does not work multithreaded so force it off.
	multithreaded = 0;

	// Tell the hddvdfs its mountpoint
	fs.set_mountpoint(mountpoint);

	if (multithreaded)
	    res = fuse_loop_mt(fuse);
	else
	    res = fuse_loop(fuse);

	fuse_teardown(fuse, mountpoint);
	if (res == -1)
	    return 1;

	fs.unmount();

	return 0;
}
