/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
I'd like to extend a special thank you to muslix64 and arnezami for
figuring out the basics of AACS. Without them this probably wouldn't have
been made. And offcourse the people who worked with them in getting through it.
*/

#include "aacs.hpp"
#include "aacs_crypto.hpp"
#include "mediakeyblock.hpp"
#include "filesystem.hpp"
#include "file.hpp"
#include "util.hpp"
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>

/**
 *
 */
AACS::AACS(Logger*				pLog,
		   const bool 			useDriveAuthentication,
		   const std::string& 	device,
		   const std::string& 	mountPoint)
: 	m_pLog(pLog),
	m_pMKB(0),
	m_pCrypto(0),
	m_useDriveAuthentication(useDriveAuthentication),
	m_device(device),
	m_mountPoint(mountPoint),
	m_isBRDisk(false)
{
	// make sure TK table is empty
	memset(m_TitleKeyTable, 0x00, 64*16);

	// setup MKB stuff
	m_pMKB = new MediaKeyBlock(m_pLog);
	m_pMKB->Load(mountPoint);

	// setup crypto stuff
	m_pCrypto = new AACSCrypto(m_pLog, useDriveAuthentication, device, m_pMKB);
}

AACS::~AACS()
{
	if (m_pCrypto)	delete m_pCrypto;
	if (m_pMKB)		delete m_pMKB;
}

/**
 * Determine if we're dealing with HD DVD or BR by trying to open the MKB
 */
void AACS::DetectMedia()
{
	m_pLog->LogInfo("Detecting media");
	const std::string 	HDDVDFilename 	= m_mountPoint + "AACS" + SEPERATOR + "MKBROM.AACS";
	const std::string 	BRFilename 		= m_mountPoint + "AACS" + SEPERATOR + "MKB_RO.inf";
	std::ifstream 		MKBFile;

	// try HD DVD
	MKBFile.open(HDDVDFilename.c_str(), std::ios::in);
	if (MKBFile.is_open())
	{
		m_pLog->LogInfo("\tDetected HD DVD");
		m_isBRDisk = false;
		return;
	}

	// try BR
	MKBFile.open(BRFilename.c_str(), std::ios::in);
	if (MKBFile.is_open())
	{
		m_pLog->LogNotice("\tDetected Blu-Ray (not supported!)");
		m_isBRDisk = true;
		return;
	}

	// neither
	throw std::runtime_error("Unable to determine disk format. Are you sure you've got a disc inserted?");
}

/**
 * Load Title Key file
 */
void AACS::LoadTitleKeyFile()
{
	m_pLog->LogInfo("Loading Title Key file");
	std::ifstream 	TKFile;
	std::string		filename;

	if (m_isBRDisk) {
		filename = m_mountPoint + "AACS" + SEPERATOR + "Unit_Key_RO.inf";
	}
	else
	{
		filename = m_mountPoint + "AACS" + SEPERATOR + "VTKF.AACS";
		TKFile.open(filename.c_str(), std::ios::binary | std::ios::in);
		TKFile.close();
		if (!TKFile.is_open()) {
			TKFile.clear();
			filename = m_mountPoint + "AACS" + SEPERATOR + "VTKF000.AACS";
		}
	}

	TKFile.open(filename.c_str(), std::ios::binary | std::ios::in);
	if (!TKFile.is_open()) {
		throw std::runtime_error("Error opening Title Key/Unit Key File: " + filename);
	}

	File oTKFile(filename);
	size_t TKFileSize = oTKFile.FileSize();

	// load TKFile into memory and extract keys
	try
	{
		uint8_t* pTitleKeyFileBuffer = oTKFile.LoadFileIntoMemory();

		if (m_isBRDisk)
		{
			/** @todo Implement BR unit key getting? */
			//GetEncryptedUnitKeys(pTitleKeyFileBuffer, TKFileSize, TitleKeys);
		}
		else
		{
			ExtractTitleKeys(pTitleKeyFileBuffer, TKFileSize);
			/** @todo Implement TKfile MAC */
			//GetTitleKeyFileMAC(pTitleKeyFileBuffer, TKFileSize, TitleKeyFileMAC);
		}
	}
	catch (std::ios_base::failure& e)
	{
		std::stringstream errorMessage;
		errorMessage << "Error loading Title Key File: " << e.what();
		throw std::runtime_error(errorMessage.str());
	}
	catch (...) {
		throw;	// pass on
	}
}

/**
 * Get encrypted Title Keys
 *
 * There are 64 possible key per TK file
 */
void AACS::ExtractTitleKeys(uint8_t* TitleKeyFileBuffer, size_t TKFileSize)
{
	m_pLog->LogInfo("Extracting encrypted Title Keys:");

	size_t 	pos 		= 128;							// skip header
	uint8_t BIFOFound	= TitleKeyFileBuffer[pos];
	int 	keyNo		= 0;
	Utils	Util;
	uint8_t TitleKey[TK_LENGTH];

	// only do valid keys (AV_FLG)
	while ((pos < TKFileSize) && ((BIFOFound & 0x80) == 0x80))
	{
		// extract key
		memcpy(TitleKey, TitleKeyFileBuffer + pos + 4, TK_LENGTH);

		// add to lookup table
		for (unsigned char i = 0; i < TK_LENGTH; i++) {
			m_TitleKeyTable[keyNo][i] = TitleKey[i];
		}

		std::stringstream message;
		message << "Encrypted Title Key " << keyNo << ":\t" << Util.ToHexString(TitleKey, 16);
		m_pLog->LogInfo(message.str());

		keyNo++;
		pos			+= 36;
		BIFOFound	 = TitleKeyFileBuffer[pos];
	}
}

/**
 * Decrypt Title Keys
 */
void AACS::DecryptTitleKeys(const std::string& strVolumeKey)
{
	if (!m_useDriveAuthentication)
	{
		uint8_t u8VolumeKey[VK_LENGTH] = { 0 };
		Utils::HexStringToByteArray(strVolumeKey, u8VolumeKey);
		m_pCrypto->setVolumeKey(u8VolumeKey);
	}
	m_pCrypto->DecryptTitleKeys(&m_TitleKeyTable[0][0]);
}

/**
 * Dumps entire TitleKey list to console
 */
void AACS::PrintTitleKeys()
{
/*
	std::cout << "TitleKey list:" << std::endl;
	std::map<int, Key>::iterator iter;
	for (iter = m_mTitleKeys.begin(); iter != m_mTitleKeys.end(); iter++)
	{
		std::cout << iter->first << " = " << iter->second.AsString() << std::endl;
	}
*/
}

/**
 * Get Title Key number from EVOB header
 * The first 128 bytes of EVOB pack is the unencrypted header.
 * Byte 0x3e: simply says "Data defined in the HD DVD-Video specification"... grmbl
 */
int AACS::GetKeyFromEVOB(const std::string EVOBFilename)
{
	m_pLog->LogInfo("Getting key number from " + EVOBFilename);

	/** @test Why can't new be used (wrong value is returned otherwise) */
	uint8_t header[128];

	std::ifstream EVOBFile(EVOBFilename.c_str(), std::ios::binary);
	if (!EVOBFile.is_open())
	{
		std::stringstream errorMessage;
		errorMessage << "Error accessing " << EVOBFilename << " while trying to get key!";
		throw std::runtime_error(errorMessage.str());
	}

	EVOBFile.read((char*)header, sizeof(header));
	return header[0x3E] - 1;	// note: we start at 0, while keynumber starts at 1
}

/**
 * Verify we have the needed (decrypted) Title Key
 */
bool AACS::HasKey(const int keyNumber)
{
	uint8_t noKey[16] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	// verify key is present by checking if it's not all zero's
	return (memcmp(&m_TitleKeyTable[keyNumber][0], noKey, TK_LENGTH) != 0);
}

//============================================================================
// Accessor junk
//============================================================================
void AACS::TitleKey(const uint8_t keyNumber, const uint8_t* TitleKey)
{
	for (int i = 0; i < 16; i++) {
		m_TitleKeyTable[keyNumber][i] = TitleKey[i];
	}
}

std::string AACS::MountPoint() const
{
	return m_mountPoint;
}

const uint8_t* AACS::TitleKeyTable() const
{
	return &m_TitleKeyTable[0][0];
}
