/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "aacs.hpp"				// key sizes
#include "aacs_crypto.hpp"
#include "hddrive.hpp"
#include "rijndael.hpp"
#include "ecdsa.hpp"
#include "util.hpp"
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <iomanip>
#include <openssl/evp.h>	//ecdsa
#include <openssl/ecdsa.h>	//ecdsa
#include <openssl/err.h>	//ecdsa

//extern Logger *Log;

/**
 *
 */
AACSCrypto::AACSCrypto(Logger* 				pLog,
					   const bool 			useDriveAuthentication,
					   const std::string& 	device,
					   MediaKeyBlock* 		pMKB)
:	m_pLog(pLog),
	m_useDriveAuthentication(useDriveAuthentication),
	m_device(device),
	m_pMKB(pMKB),
	m_VolumeUniqueKeyIsValid(false),
	m_MediaKeyIsValid(false),
	m_VolumeIDIsValid(false),
	m_ProcessingKeyIsValid(false)
{
	// clear keys
	memset(m_MediaKey, 			0x00, sizeof(m_MediaKey));
	memset(m_VolumeID, 			0x00, sizeof(m_VolumeID));
	memset(m_DeviceKey, 		0x00, sizeof(m_DeviceKey));
	memset(m_MediaKeyData, 		0x00, sizeof(m_MediaKeyData));
	memset(m_VolumeUniqueKey, 	0x00, sizeof(m_VolumeUniqueKey));

	// Load processing keys
	// processing key based on MKB version
	/** @todo Find a better solution for this */
	//std::string strProcessingKeyFile;
	uint8_t ProcessingKey_v1[16] = { 0x09, 0xf9, 0x11, 0x02, 0x9d, 0x74, 0xe3, 0x5b, 0xd8, 0x41, 0x56, 0xc5, 0x63, 0x56, 0x88, 0xc0 };
	uint8_t ProcessingKey_v3[16] = { 0x45, 0x5f, 0xe1, 0x04, 0x22, 0xca, 0x29, 0xc4, 0x93, 0x3f, 0x95, 0x05, 0x2b, 0x79, 0x2a, 0xb2 };
	if (pMKB->Version() == 0x00000001) {
		//strProcessingKeyFile = "keys/processingkey_v1.hex";
		//09 f9 11 02 9d 74 e3 5b d8 41 56 c5 63 56 88 c0
		memcpy(m_ProcessingKey, ProcessingKey_v1, 16);
	}
	else if (pMKB->Version() == 0x00000003) {
		//strProcessingKeyFile = "keys/processingkey_v3.hex";
		//45 5f e1 04 22 ca 29 c4 93 3f 95 05 2b 79 2a b2
		memcpy(m_ProcessingKey, ProcessingKey_v3, 16);
	}
	/*
	m_pLog->LogInfo("Loading processing key from " + strProcessingKeyFile);
	std::ifstream ifProcessingKey(strProcessingKeyFile.c_str(), std::ios::binary);
	if (!ifProcessingKey.is_open())
	{
		throw std::runtime_error("Error loading processing key.");
	}
	ifProcessingKey.rdbuf()->sgetn((char*)m_ProcessingKey, 16);
	ifProcessingKey.close();
	*/
	m_ProcessingKeyIsValid = true;

	/** @todo Extract the UV number from MKB and calc instead of hardcoding */
	switch (pMKB->Version())
	{
		case 1L:
			m_UVnumber[0] = 0x00;
			m_UVnumber[1] = 0x00;
			m_UVnumber[2] = 0x00;
			m_UVnumber[3] = 0x01;
			break;
		case 3L:
			m_UVnumber[0] = 0x00;
			m_UVnumber[1] = 0x00;
			m_UVnumber[2] = 0x00;
			m_UVnumber[3] = 0x49;
			break;
	}
}

/**
 * AES ECB decryption
 * @see AACS spec common 2.1.1
 */
void AACSCrypto::AES_128D(uint8_t* k, uint8_t* d, uint8_t* result)
{
	CRijndael oRijndael;
	oRijndael.MakeKey((char*)k, CRijndael::sm_chain0, 16, 16);
	oRijndael.DecryptBlock((char*)d, (char*)result);
}

/**
 * AES 128 encryption
 */
void AACSCrypto::AES_128E(uint8_t* k, uint8_t* d, uint8_t* result)
{
	CRijndael oRijndael;
	oRijndael.MakeKey((char*)k, CRijndael::sm_chain0, 16, 16);
	oRijndael.EncryptBlock((char*)d, (char*)result);
}

/**
 * AES CBC decryption
 * @see AACS spec common 2.1.2
 */
void AACSCrypto::AES_128CBCD(uint8_t* k, uint8_t* d, uint8_t* result)
{
	const unsigned char AESCBCConstantIV[] = {
		0x0B, 0xA0, 0xF8, 0xDD, 0xFE, 0xA6, 0x1F, 0xB3,
		0xD8, 0xDF, 0x9F, 0x56, 0x6A, 0x05, 0x0F, 0x78
	};

	CRijndael oRijndael;
	oRijndael.MakeKey((char*)k, (char*)AESCBCConstantIV, 16, 16);
	//oRijndael.Decrypt((char*)d, result, uiEncryptedSize, CRijndael::CBC);
	oRijndael.Decrypt((char*)d, (char*)result, CRijndael::CBC);
}

/**
 * AES based one-way function (AES-128D with XOR)
 *
 * @param x1		Decryption key
 * @param x2		Data to be decrypted
 * @param result	Decrypted result
 *
 * @see AACS spec common 2.1.3
 */
void AACSCrypto::AES_G(uint8_t* x1, uint8_t* x2, uint8_t* result)
{
	CRijndael oRijndael;
	oRijndael.MakeKey((char*)x1, CRijndael::sm_chain0, 16, 16);
	oRijndael.DecryptBlock((char*)x2, (char*)result);
	Xor(result, x2);
}

/**
 * Triple AES decryption
 *
 * @param k			(Subsidiary) Device Key
 * @param s			Step (0=left child sub dev key, 1=device key, 2=right child sub dev key)
 * @param result	Decrypted result
 *
 * @see AACS spec common 3.2.2
 * Calc Subsidiary Device Keys and Processing Keys:
 * Executed three times to produce 384 output bits, incrementing the seed register by one each time. The output
 * of AES-128D is XORed with the seed registers output at each step. For each AES-G3 calculation, the seed
 * register is initialized.
 */
void AACSCrypto::AES_G3(uint8_t* k, uint8_t* result, uint8_t step)
{
	// AES-G3 seed register init
	uint8_t ucSeed[16] = {
		0x7B,0x10,0x3C,0x5D,0xCB,0x08,0xC4,0xE5,0x1A,0x27,0xB0,0x17,0x99,0x05,0x3B,0xD9
	};
	ucSeed[15] += step;

	CRijndael oRijndael;
	oRijndael.MakeKey((char*)k, CRijndael::sm_chain0, 16, 16);
	oRijndael.DecryptBlock((char*)ucSeed, (char*)result);
	Xor(result, ucSeed);
}

/**
 * AES CMAC
 */
void AACSCrypto::AES_CMAC(uint8_t* key, uint8_t* input, int length, uint8_t* MAC)
{
	uint8_t	X[16], Y[16], M_last[16], padded[16];
	uint8_t	K1[16], K2[16];
	int		n, i, flag;

	GenerateSubkey(key, K1, K2);

	n = (length + 15) / 16;      		// n is number of rounds

	if (n == 0) {
		n	 = 1;
		flag = 0;
	}
	else
	{
		if ((length % 16) == 0) { 		// last block is a complete block
			flag = 1;
		}
		else { 							// last block is not complete block
			flag = 0;
		}
	}

	if (flag) { 						// last block is complete block
		Xor(&input[16 * (n - 1)], K1, M_last);
	}
	else {
		Padding(&input[16 * (n - 1)], padded, length % 16);
		Xor(padded, K2, M_last);
	}

	for (i = 0; i < 16; i++) {
		X[i] = 0;
	}

	for (i = 0; i < n - 1; i++)
	{
		Xor(X, &input[16 * i], Y); 		// Y := Mi (+) X
		AES_128E(key, Y, X);			// X := AES-128(KEY, Y);
	}

	Xor(X, M_last, Y);
	AES_128E(key, Y, X);

	for (i = 0; i < 16; i++) {
		MAC[i] = X[i];
	}
}

/**
 * AES CMAC bitshifter
 */
void AACSCrypto::LeftShiftOneBit(uint8_t* input, uint8_t* output)
{
	int		i;
	uint8_t	overflow = 0;

	for (i = 15; i >= 0; i--)
	{
		output[i]	 = input[i] << 1;
		output[i]	|= overflow;
		overflow	 = (input[i] & 0x80) ? 1 : 0;
	}
}

/**
 * AES CMAC subkey generation
 */
void AACSCrypto::GenerateSubkey(uint8_t* key, uint8_t* K1, uint8_t* K2)
{
	uint8_t L[16];
	uint8_t Z[16];
	uint8_t tmp[16];
	int		i;
	uint8_t const_Rb[16] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x87
	};

	for (i = 0; i < 16; i++) {
		Z[i] = 0;
	}

	AES_128E(key, Z, L);

	if ((L[0] & 0x80) == 0) {			// If MSB(L) = 0, then K1 = L << 1
		LeftShiftOneBit(L, K1);
	}
	else {								// Else K1 = ( L << 1 ) (+) Rb
		LeftShiftOneBit(L, tmp);
		Xor(tmp, const_Rb, K1);
	}

	if ((K1[0] & 0x80) == 0) {
		LeftShiftOneBit(K1, K2);
	}
	else {
		LeftShiftOneBit(K1, tmp);
		Xor(tmp, const_Rb, K2);
	}
}

/**
 * AES CMAC padding
 */
void AACSCrypto::Padding(uint8_t* lastb, uint8_t* pad, int length)
{
	int j;

	// original last block
	for (j = 0; j < 16; j++)
	{
		if (j < length)			pad[j] = lastb[j];
		else if (j == length)	pad[j] = 0x80;
		else					pad[j] = 0x00;
	}
}

/**
 *
 */
bool AACSCrypto::AACS_Sign(	uint8_t* 			Certificate,
							const std::string&	PrivateKeyStr,
							uint8_t*			Signature,
							uint8_t* 			Nonce,
							uint8_t* 			Point)
{
	EC_KEY		*pKey = Key();
	BIGNUM		*PrivateKey(0);
	ECDSA_SIG	*pSig(0);

	if (pKey)
	{
		EC_KEY_auto_ptr apKey(pKey);

		if (SetCertificate(pKey, Certificate))
		{
			if (BN_hex2bn(&PrivateKey, PrivateKeyStr.c_str()) && EC_KEY_set_private_key(pKey, PrivateKey))
			{
				uint8_t 		md_value[EVP_MAX_MD_SIZE];
				unsigned int	md_len;
				EVP_MD_CTX 		mdctx;

				EVP_MD_CTX_init(&mdctx);							// init digest
				EVP_DigestInit(&mdctx, EVP_ecdsa());				// setup digest context
				EVP_DigestUpdate(&mdctx, Nonce, 20);				// hash nonce
				EVP_DigestUpdate(&mdctx, Point, 40);				// hash point
				EVP_DigestFinal_ex(&mdctx, md_value, &md_len);		// retrieve digest

				// compute the ECDSA signature
				pSig = ECDSA_do_sign(md_value, md_len, pKey);
				ECDSA_SIG_auto_ptr apSig(pSig);

				if (BN_bn2bin(pSig->r, Signature) == 20) {
					if (BN_bn2bin(pSig->s, Signature + 20) == 20) {
						return pSig;
					}
				}
			}
		}
	}

	return false;
}

/**
 * Verify signature is valid
 */
bool AACSCrypto::AACS_Verify(	uint8_t* Certificate,
								uint8_t* Signature,
								uint8_t* Nonce,
								uint8_t* Point)
{
	// get EC_KEY object
	EC_KEY*	pKey = Key();

	if (pKey)
	{
		EC_KEY_auto_ptr apKey(pKey);

		if (SetCertificate(pKey, Certificate))
		{
			uint8_t 		md_value[EVP_MAX_MD_SIZE];
			unsigned int 	md_len;
			EVP_MD_CTX 		mdctx;
			ECDSA_SIG		*pSig;

			EVP_MD_CTX_init(&mdctx);						// init digest
			EVP_DigestInit(&mdctx, EVP_ecdsa());			// setup digest context
			EVP_DigestUpdate(&mdctx, Nonce, 20);			// hash nonce
			EVP_DigestUpdate(&mdctx, Point, 40);			// hash point
			EVP_DigestFinal_ex(&mdctx, md_value, &md_len);	// retrieve digest

			// create ECDSA_SIG and set r and s values
			pSig = ECDSA_SIG_new();

			if (pSig)
			{
				ECDSA_SIG_auto_ptr apSig(pSig);

				if (BN_bin2bn(Signature, 20, pSig->r) && BN_bin2bn(Signature + 20, 20, pSig->s))
				{
					// verify signature is valid
					if (ECDSA_do_verify(md_value, md_len, pSig, pKey) > 0) {
						return true;
					}
				}
			}
		}
	}

	return false;
}

/**
 * Calculate Bus Key
 *
 * @param HostCert		Host certificate
 * @param HostKey		Host key
 * @param DriveKeyPoint	Drive key point
 * @param BusKey		Resulting Bus Key
 *
 * @see AACS spec common, chapter 4
 * 29. The host calculates the Bus Key from a point of the elliptic curve as follows
 *	BK = [x-coordinate of Hk Dv]lsb_128
 *	where BK is the least significant 128-bit of x-coordinate.
 */
bool AACSCrypto::CalculateBusKey(	uint8_t* HostCertificate,
									uint8_t* HostKey,
									uint8_t* DriveKeyPoint_bin,
									uint8_t* BusKey)
{
	EC_KEY			*pKey			= Key();
	BIGNUM			*pHostKey_bn	= BN_bin2bn(HostKey, 20, NULL);
	BIGNUM			*x				= BN_bin2bn(DriveKeyPoint_bin, 20, NULL);
	BIGNUM			*y				= BN_bin2bn(DriveKeyPoint_bin + 20, 20, NULL);
	BN_CTX			*pCtx(0);
	EC_POINT		*pDriveKeyPoint_pnt(0);
	EC_POINT		*pBusKeyPoint(0);
	const EC_GROUP	*pGroup(0);
	uint8_t 		busKeyLong[100];
	BIGNUM			*pBusKeyPoint_bn(0);

	BN_auto_ptr apX(x);
	BN_auto_ptr apY(y);

	if (pKey && pHostKey_bn && x && y) {
		if (SetCertificate(pKey, HostCertificate)) {
			pGroup = EC_KEY_get0_group(pKey);
		}
	}
	else {
		return false;
	}

	if (pGroup) {
		if (pDriveKeyPoint_pnt = EC_POINT_new(pGroup)) {
			if (EC_POINT_set_affine_coordinates_GFp(pGroup, pDriveKeyPoint_pnt, x, y, NULL)) {
				if (pDriveKeyPoint_pnt) {
					if (pBusKeyPoint = EC_POINT_new(pGroup)) {
						if (EC_POINT_mul(pGroup, pBusKeyPoint, NULL, pDriveKeyPoint_pnt, pHostKey_bn, NULL)) {
							if (pBusKeyPoint_bn = EC_POINT_point2bn(pGroup, pBusKeyPoint, POINT_CONVERSION_UNCOMPRESSED, NULL, NULL)) {
								if (BN_bn2bin(pBusKeyPoint_bn, busKeyLong))
								{
									// copy to Bus Key
									memcpy(BusKey, busKeyLong + 4 + 1, 16); 	// +1 due to uncompressed code 0x04
									//BN_clear_free(x);
									//BN_clear_free(y);
									return true;
								}
							}
						}
					}
				}
			}
		}
	}

	return false;
}


/**
 * XOR buffer of 16 bytes
 */
void AACSCrypto::Xor(uint8_t* buff, uint8_t const* chain)
{
	int i;

	for (i = 0; i < 16; i++) {
		buff[i] ^= *(chain++);
	}
}

void AACSCrypto::Xor(uint8_t* a, uint8_t* b, uint8_t* out)
{
	int i;

	for (i = 0; i < 16; i++) {
		out[i] = a[i] ^ b[i];
	}
}

/**
 * Create EC_KEY object
 */
EC_KEY* AACSCrypto::Key()
{
	// make key
	EC_KEY* pKey = EC_KEY_new();
	if (!pKey) {
		return 0;
	}

	// assign group
	EC_GROUP *pGroup = AACS_group();
	if (!EC_KEY_set_group(pKey, pGroup))
	{
		EC_KEY_free(pKey);
		return 0;
	}

	return pKey;
}

/**
 * Set ECDSA certificate
 */
int AACSCrypto::SetCertificate(EC_KEY* pKey, uint8_t* pCertificate)
{
	EC_POINT		*pPublicKey(0);
	BIGNUM			*x = BN_bin2bn(pCertificate + 12, 20, NULL);
	BIGNUM			*y = BN_bin2bn(pCertificate + 32, 20, NULL);
	const EC_GROUP 	*pGroup = EC_KEY_get0_group(pKey);

	BN_auto_ptr apX(x);
	BN_auto_ptr apY(y);

	if ((pCertificate[1] == 0) && (*(short *)(pCertificate + 2) == 0x5c00)) {
		if (x && y) {
			if (pPublicKey = EC_POINT_new(pGroup)) {
				if (EC_POINT_set_affine_coordinates_GFp(pGroup, pPublicKey, x, y, NULL)) {
					return EC_KEY_set_public_key(pKey, pPublicKey);
				}
			}
		}
	}

	return 0;
}

/**
 * Create ECDSA EC_GROUP
 *
 * Has AACS ECC parameters
 */
static EC_GROUP* AACS_group()
{
	EC_GROUP	*pGroup(0);
	EC_POINT	*pPoint(0);
	BN_CTX		*ctx(0);
	BIGNUM		*p(0), *a(0), *b(0), *x(0), *y(0), *order(0);

	// ECC parameters
	/** @see AACS spec common, chapter 2.3 */
	const char data_p[] = "900812823637587646514106462588455890498729007071";
	const char data_a[] = "-3";
	const char data_b[] = "366394034647231750324370400222002566844354703832";
	const char data_x[] = "264865613959729647018113670854605162895977008838";
	const char data_y[] = "51841075954883162510413392745168936296187808697";
	const char data_r[] = "900812823637587646514106555566573588779770753047";

	// alloc mem
	ctx 	= BN_CTX_new();	BN_CTX_auto_ptr apCtx(ctx);
	p 		= BN_new();		BN_auto_ptr apP(p);
	a 		= BN_new();		BN_auto_ptr apA(a);
	b 		= BN_new();		BN_auto_ptr apB(b);
	x 		= BN_new();		BN_auto_ptr apX(x);
	y		= BN_new();		BN_auto_ptr apY(y);
	order	= BN_new();		BN_auto_ptr apOrder(order);

	if (!ctx || !p || !a || !b || !x || !y || !order)
	{
		ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_MALLOC_FAILURE);
		return EC_GROUP_NULL;
	}

	//
	if (BN_dec2bn(&p, data_p) && BN_dec2bn(&a, data_a) && BN_dec2bn(&b, data_b))		// convert ECC params to BIGNUMs
	{
		pGroup	= EC_GROUP_new_curve_GFp(p, a, b, ctx); 				// new curve group
		//EC_GROUP_auto_ptr apGroup(pGroup);

		if (pGroup)
		{
			pPoint = EC_POINT_new(pGroup); 								// new curve point
			EC_POINT_auto_ptr apPoint(pPoint);

			if (pPoint)
				if (BN_dec2bn(&x, data_x) && BN_dec2bn(&y, data_y))
					if (EC_POINT_set_affine_coordinates_GF2m(pGroup, pPoint, x, y, ctx))
						if (BN_dec2bn(&order, data_r) && BN_set_word(x, 1))
							if (EC_GROUP_set_generator(pGroup, pPoint, order, x))
								//return apGroup.release();
								return pGroup;
							else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_EC_LIB);
						else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_BN_LIB);
					else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_EC_LIB);
				else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_BN_LIB);
			else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_EC_LIB);
		}
		else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_EC_LIB);
	}
	else ECerr(EC_F_EC_GROUP_NEW_FROM_DATA, ERR_R_BN_LIB);

	return EC_GROUP_NULL;
}

/**
 * Calculate Volume ID MAC
 *
 * Used by HDDrive to verify validity of the Volume ID
 *
 * @param VolumeID		Volume ID
 * @param length
 * @param BusKey		Bus Key
 * @param VolumeIDMAC	Resulting Volume ID MAC
 *
 * @see AACS spec common
 * 5. The host calculates a message authentication code (Hm) from the Volume_ID received in step 4 and the Bus Key calculated in step 27
 *	Hm = CMAC(BK, Volume_ID)
 */
bool AACSCrypto::VolumeIDIsValid (uint8_t* BusKey, uint8_t* VolumeID, uint8_t* VolumeIDMAC)
{
	uint8_t CMAC[16];
	memset(CMAC, 0x00, sizeof(CMAC));

	AES_CMAC(BusKey, VolumeID, 16, CMAC);

	return (memcmp(CMAC, VolumeIDMAC, 16) == 0);
}

/**
 * Calc VUK (AES-G)
 *
 * @see AACS spec prerecorded
 *
 * The Volume Unique Key (Kvu) and/or the Volume Variant Unique Keys (Kvvu ) are used to encrypt and decrypt
 * the Title Keys stored on the pre-recorded media.
 * The Volume Unique Key is calculated as follows:
 * 	1. Calculate the Media Key (Km)
 *	2. Calculate the Volume Unique Key (Kvu):
 *		Kvu = AES-G(Km, Volume ID)
 */
void AACSCrypto::CalculateVolumeUniqueKey()
{
	if (!m_MediaKeyIsValid) {
		CalculateMediaKey();
	}

	HDDrive Drive(m_pLog, m_device, this);
	if (!Drive.IsAuthenticated())
	{
		Drive.Authenticate();
		m_VolumeIDIsValid = true;
	}

	m_pLog->LogInfo("Calculating Volume Unique Key:");
	AES_G(m_MediaKey, Drive.VolumeID(), m_VolumeUniqueKey);

	m_pLog->LogInfo("\tVolume Unique Key\t" + Utils::ToHexString(m_VolumeUniqueKey, 16));
	m_VolumeUniqueKeyIsValid = true;
}

/**
 * Decrypt Media Key Data (aka C-value) and make Media Key
 * Once the device has the correct Device Key D, it calculates a Processing Key K using AES-G3
 * Using that Processing Key K and the appropriate 16 bytes of encrypted key data C, the device calculates
 * the 128-bit Media Key Km as follows:
 *
 * 	Km = AES-128D(K, C) XOR (0x000000000000000000000000 || uv)
 */
void AACSCrypto::CalculateMediaKey()
{
	/** @todo CalculateProcessingKey()? Right now it's hardcoded */
	/*
	if (!m_bProcessingKeyIsValid)
	{
		CalculateProcessingKey();
	}
	*/

	int iRecordNo(0);
	switch (m_pMKB->Version())
	{
		case 1L:
			iRecordNo = 0;
			break;
		case 3L:
			iRecordNo = 4;
			break;
	}

	m_pLog->LogInfo("Calculating Media Key:");

	// decrypt Media Key Data using the Processing Key
	uint8_t MediaKeyDataDec[16];
	AES_128D(m_ProcessingKey, m_pMKB->MediaKeyData(iRecordNo), MediaKeyDataDec);

	// XOR and glue
	// UV number is only 4 bytes, so leave 12 MSB bytes untouched
	uint8_t i;
	for (i = 0; i < 16; i++)
	{
		if (i < 12)	m_MediaKey[i] = MediaKeyDataDec[i];
		else		m_MediaKey[i] = MediaKeyDataDec[i] ^ m_UVnumber[i - 12];
	}

	std::stringstream ssMsg;
	ssMsg << "\tProcessing Key\t" << Utils::ToHexString(m_ProcessingKey, 16);
	m_pLog->LogInfo(ssMsg.str());
	m_pLog->LogInfo("\tMedia Key Data\t" + Utils::ToHexString(m_pMKB->MediaKeyData(iRecordNo), 16));
	m_pLog->LogInfo("\tDecrypt result\t" + Utils::ToHexString(MediaKeyDataDec, 16));
	m_pLog->LogInfo("\tMedia Key\t"		 + Utils::ToHexString(m_MediaKey, 16));

	VerifyMediaKey();
	m_MediaKeyIsValid = true;
}

/**
 * Decrypt Verification Data
 *
 * Needs Media Key
 */
void AACSCrypto::VerifyMediaKey()
{
	m_pLog->LogInfo("Verifying Media Key");

	uint8_t u8VerificationData[8] =	{ 0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF };
	uint8_t u8DecryptedVerificationData[16];

	// get encrypted data from MKB
	AES_128D(m_MediaKey, m_pMKB->VerificationData(), u8DecryptedVerificationData);

	m_pLog->LogInfo("\tVerify Data\t" 	+ Utils::ToHexString(m_pMKB->VerificationData(), 16));
	m_pLog->LogInfo("\tVerify result\t" + Utils::ToHexString(u8DecryptedVerificationData, 8));

	if (!(memcmp(u8VerificationData, u8DecryptedVerificationData, 8) == 0)) {
		throw std::runtime_error("Media Key verification failed!");
	}

	m_pLog->LogInfo("\tMedia Key is valid");
}

/**
 * Decrypt Title Keys (AES-128D)
 *
 * This will do drive authentication if needed. If no authentication is used, the VUK should be set first
 */
void AACSCrypto::DecryptTitleKeys(uint8_t* TitleKeyTable)
{
	m_pLog->LogInfo("Decrypting Title Keys");

	// start doing drive authentication if needed
	if (m_useDriveAuthentication)
	{
		m_pLog->LogInfo("\tUsing drive authentication");
		// check if we have the VUK already
		if (!m_VolumeUniqueKeyIsValid) {
			CalculateVolumeUniqueKey();
		}
	}

	// use user set key
	// key should be set first
	else
	{
		// check if we have the VUK
		if (!m_VolumeUniqueKeyIsValid) {
			throw std::runtime_error("Volume Unique Key not set");
		}

		// check if VUK contains anything meaningful
		uint8_t u8EmptyKey[VK_LENGTH] = { 0 };
		if ((memcmp(m_VolumeUniqueKey, u8EmptyKey, VK_LENGTH) == 0)) {
			throw std::runtime_error("Volume Unique Key not set");
		}
	}

	m_pLog->LogDebug(Utils::ToHexString(m_VolumeUniqueKey, TK_LENGTH));
	uint8_t u8DecryptedKey[TK_LENGTH];
	int		iKeyNo;

	for (iKeyNo = 0; iKeyNo < 64; iKeyNo++)
	{
		AES_128D(m_VolumeUniqueKey, TitleKeyTable + iKeyNo * TK_LENGTH, u8DecryptedKey);
		memcpy(TitleKeyTable + iKeyNo * TK_LENGTH, u8DecryptedKey, sizeof(u8DecryptedKey));

		std::stringstream ssMsg;
		ssMsg << "Decrypted Title Key " << iKeyNo << ":\t" << Utils::ToHexString(u8DecryptedKey, TK_LENGTH);
		m_pLog->LogInfo(ssMsg.str());
	}
}

//------------------------ Accessor junk -----------------------------------------------

/**
 * Set VUK
 * @note Sets VUK validity flag to true
 */
void AACSCrypto::setVolumeKey(const uint8_t* const u8Key)
{
	memcpy(m_VolumeUniqueKey, u8Key, VK_LENGTH);
	m_VolumeUniqueKeyIsValid = true;
}

bool AACSCrypto::isVolumeUniqueKeyValid()
{
	return m_VolumeUniqueKeyIsValid;
}
