/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef AACS_CRYPTO_H
#define AACS_CRYPTO_H

#include "mediakeyblock.hpp"
#include "logger.hpp"
#include <openssl/ec.h>
#include <stdint.h>

static EC_GROUP *AACS_group();

class AACSCrypto
{
public:
	AACSCrypto(Logger* 				pLog,
			   const bool 			useDriveAuthentication,
			   const std::string&	device,
			   MediaKeyBlock* 		pMKB);

	void	AES_128D					(uint8_t*			k,
										 uint8_t*			d,
										 uint8_t*			result
										);
	void	AES_128E					(uint8_t* 			k,
										 uint8_t* 			d,
										 uint8_t* result
										);
	void	AES_128CBCD					(uint8_t*			k,
										 uint8_t*			d,
										 uint8_t*			result
										);
	void	AES_G						(uint8_t*			x1,
										 uint8_t*			x2,
										 uint8_t*			result
										);
	void 	AES_G3						(uint8_t* 			k,
										 uint8_t* 			result,
										 uint8_t 			step
										);
	void	AES_CMAC					(uint8_t*			key,
										 uint8_t*			input,
										 int 				length,
										 uint8_t*			mac
										);
	bool 	AACS_Sign					(uint8_t* 			Certificate,
										 const std::string&	PrivateKeyStr,
										 uint8_t*			Signature,
										 uint8_t* 			Nonce,
										 uint8_t* 			Point
										);
	bool 	AACS_Verify					(uint8_t* 			Certificate,
										 uint8_t* 			Signature,
										 uint8_t* 			Nonce,
										 uint8_t* 			Point
										);
	bool 	CalculateBusKey				(uint8_t* 			HostCertificate,
										 uint8_t* 			HostKey,
										 uint8_t* 			DriveKeyPoint,
										 uint8_t* 			BusKey
										);
	bool 	VolumeIDIsValid				(uint8_t* 			BusKey,
										 uint8_t* 			VolumeID,
										 uint8_t* 			VolumeIDMAC
										);
	void	CalculateVolumeUniqueKey	();
	void	CalculateMediaKey			();
	void	VerifyMediaKey				();
	void	DecryptTitleKeys			(uint8_t* 			TitleKeyTable);

	// accessors
	void	setVolumeKey(const uint8_t* const u8Key);
	bool	isVolumeUniqueKeyValid();

private:
	Logger*			m_pLog;
	bool			m_useDriveAuthentication;
	std::string		m_device;
	MediaKeyBlock*	m_pMKB;

	void 	Xor				(uint8_t* buff, uint8_t const* chain);
	void	Xor				(uint8_t* a, uint8_t* b, uint8_t* out);
	void	LeftShiftOneBit	(uint8_t* input, uint8_t* output);
	void	GenerateSubkey	(uint8_t* key, uint8_t* K1, uint8_t* K2);
	void	Padding			(uint8_t* lastb, uint8_t* pad, int length );
	int 	SetCertificate	(EC_KEY* pKey, uint8_t* Certificate);
	EC_KEY* Key				();

	// validity flags
	bool		m_VolumeUniqueKeyIsValid;
	bool		m_MediaKeyIsValid;
	bool		m_VolumeIDIsValid;
	bool		m_ProcessingKeyIsValid;

	// keys
	uint8_t		m_MediaKey[16];
	uint8_t		m_VolumeID[16];
	uint8_t		m_DeviceKey[16];
	uint8_t		m_ProcessingKey[16];
	uint8_t		m_MediaKeyData[16];
	uint8_t		m_UVnumber[4];
	uint8_t		m_VolumeUniqueKey[16];
};

#endif
