/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "acafile.hpp"
#include "resourcefile.hpp"
#include "crc32.hpp"
#include "filesystem.hpp"
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <exception>
#include <algorithm>	// transform
#include <cctype>		// tolower
#include <stdexcept>

/**
 *
 */
ACAFile::ACAFile(Logger* pLog)
:	m_pLog(pLog),
	m_filesize(0),
	m_decryptedACASize(0),
	m_pDecryptedACA(0),
	m_fp(0)
{
}

ACAFile::~ACAFile()
{
	if (m_pDecryptedACA)
		delete[] m_pDecryptedACA;
}

/**
 * Calc decrypted filesize
 *
 * Probably only useful if you need to know the filesize before actually
 * decrypting it. The way encrypted ACA files are build up makes this very
 * inefficient. Avoid usage if possible.
 */
unsigned long ACAFile::CalcDecryptedFilesize(std::istream& ACAStream)
{
/*
	// only calc if it hasn't een done yet
	if (m_decryptedFilesize > 0 ) {
		return m_decryptedFilesize;
	}

	// get header first if needed
	if (m_fileList.size() == 0)
	{
		ExtractList(ACAStream);
	}

	// add size of ACA header and length of ARF headers
	// 272 + 11, only valid when AACS encrypted
	m_decryptedFilesize += 32 + m_noOfFiles * 283;

	// get size of each file (needs header of each ARF)
	std::vector<ResourceFile*>::const_iterator iter;
	for (iter = m_fileList.begin(); iter < m_fileList.end(); ++iter)
	{
		//m_decryptedFilesize += (*iter)->dataSize; // incl AACS junk!
		todo -> extract real size from each ARF header...
	}

	std::stringstream message;
	message << "Decrypted ACA filesize is " << m_decryptedFilesize;
	m_pLog->LogDebug(message.str());
*/
}

/**
 * Decrypt entire ACA file (and output seperate files and decrypted ACA file)
 */
void ACAFile::Decrypt(const std::string& 	sourceFilename,
					  const std::string& 	destinationDir,	// incl /
					  const uint8_t* const	pTitleKeyTable,
					  bool					writeToDisk)
{
	m_pLog->LogInfo("Decrypting " + sourceFilename);
	m_pLog->LogDebug(sourceFilename);
	m_pLog->LogDebug(destinationDir);

	// get ACA filename and dir
	const std::string decryptedFilename =
		sourceFilename.substr(sourceFilename.rfind('/', sourceFilename.length()) + 1);
	m_outputDir = destinationDir + decryptedFilename + "_files";

	// create dir to store seperate files, overwrite if it already exists
	if (writeToDisk)
	{
		m_pLog->LogNotice("Creating directory " + m_outputDir);
		Filesystem::CreateDir(m_outputDir, true);
	}

	// open and decrypt ARF
	std::ifstream ACAStream;
	Open(sourceFilename, pTitleKeyTable, ACAStream, writeToDisk);

	// write to disc
	if (writeToDisk)
	{
		const std::string decryptedFullPath = destinationDir + decryptedFilename;
		m_pLog->LogNotice("Saving decrypted aca to " + decryptedFullPath);

		// create dir
		Filesystem::CreateDir(destinationDir, false);

		std::ofstream decryptedFile(decryptedFullPath.c_str(), std::ios::binary);
		if (!decryptedFile.is_open())
		{
			std::stringstream errorMsg;
			errorMsg << "Error saving decrypted aca file to " << decryptedFullPath;
			throw std::runtime_error(errorMsg.str());
		}
		decryptedFile.write((char*)m_pDecryptedACA, m_fp);
		decryptedFile.close();
	}
}

/**
 * HDDVDFS
 *
 * ACA files are a pain in the a$$ concerning this HDDVDFS. The fileindex works
 * with offsets, which are only known after the file is decrypted.
 * The original file is loaded into memory and decrypted. While this may not be
 * the most elegant way, I assume that if you can play HD DVD, your computer has
 * enough memory and is fast enough. After this when calling read it'll return
 * the already decrypted data. Biggest aca files I've seen to date are just over
 * 40 MB which should be very doable considering the target systems.
 * Besides otherwise there's the risk of decrypting everytime one calls read().
 * And now there is no drive access when reading. It's read from memory and no
 * way a drive can compete with that (especially not the Xbox360 USB HD DVD
 * addon)!
 */
void ACAFile::Open(const std::string&	sourceFilename,
				   const uint8_t* const pTitleKeyTable,
				   std::ifstream&		ACAStream,
				   bool					writeToDisk)
{
	m_pLog->LogInfo("Decrypting " + sourceFilename);

	// set Title Key table
	m_pTitleKeyTable = (uint8_t*)pTitleKeyTable;

	// open file
	ACAStream.open(sourceFilename.c_str(), std::ios::binary);
	if (!ACAStream.is_open())
	{
		std::stringstream msg;
		msg << "Error opening " << sourceFilename;
		throw std::runtime_error(msg.str());
	}

	// get info on included files
	ExtractList(ACAStream);

	// the final decrypted size will be smaller due to the not present hash bytes and what not
	m_decryptedACASize = m_filesize - m_noOfFiles * (11 + 272);

	// alloc memory
	try
	{
		m_pDecryptedACA = new uint8_t[m_decryptedACASize];
		// m_decryptedACASize can now be updated with real sizeinfo as ARFs are processed

		// extract original header
		// ARF size, CRC and MIME will be replaced after decrypting
		LoadHeader(ACAStream);

		// process ARF data
		m_pLog->LogInfo("*** Processing actual ARF data ***");
		m_fp 								= m_fileList.at(0)->offset;
		m_decryptedACASize 					= m_fp;
		unsigned long offsetInOriginalACA 	= m_fp;

		int fileNo;
		for (fileNo = 0; fileNo < m_noOfFiles; fileNo++)
		{
			std::stringstream offsetMessage;
			offsetMessage << "offsetInOriginalACA = 0x" << std::hex << offsetInOriginalACA;
			m_pLog->LogDebug(offsetMessage.str());

			ARFile ARF(m_pLog, m_pTitleKeyTable);
			offsetInOriginalACA += ARF.ReadHeader(ACAStream, offsetInOriginalACA);
			offsetInOriginalACA += ARF.ReadData(ACAStream, offsetInOriginalACA);
			ARF.ExtractResource();

			// insert (decrypted) resource into final ACA buffer
			m_pLog->LogDebug("Appending decrypted data");
			memcpy(m_pDecryptedACA + m_fp, ARF.ResourceFile(), ARF.Filesize());

			// update info in header and size
			UpdateHeader(fileNo, ARF);
			m_decryptedACASize += ARF.Filesize();

			// write seperate ARF to disk
			if (writeToDisk)
			{
				const std::string outputFile = m_outputDir + "/" + m_fileList.at(fileNo)->filename;
				m_pLog->LogNotice("Writing " + outputFile);

				std::ofstream ARFFile(outputFile.c_str(), std::ios::binary);
				if (!ARFFile.is_open()) {
					throw std::runtime_error("Error writing " + outputFile);
				}

				ARFFile.write((char*)ARF.ResourceFile(), ARF.Filesize());
				ARFFile.close();
			}
		}

		ACAStream.close();

		// done processing all ARF's
		m_pLog->LogInfo("Finalising decrypted and reconstructed ACA");

		// update size
		// fp should now point to right after the last ARF
		m_pDecryptedACA[14] = (m_fp & 0xFF000000) >> 24;
		m_pDecryptedACA[15] = (m_fp & 0x00FF0000) >> 16;
		m_pDecryptedACA[16] = (m_fp & 0x0000FF00) >> 8;
		m_pDecryptedACA[17] =  m_fp & 0x000000FF;

	}
	catch (std::bad_alloc& ) {
		throw std::runtime_error("Error allocating memory for temp decrypted data storage");
	}

	// decryted ACA is constructed, now do something with it
}

/**
 * HDDVDFS
 */
unsigned long ACAFile::ReadChunk(unsigned long offset,
								 unsigned long chunkSize,
								 uint8_t* const buffer)
{
	if (offset + chunkSize > m_decryptedACASize) {
		m_pLog->LogError("Requested size is bigger than available bytes");
		return 0;
	}

	memcpy(buffer, m_pDecryptedACA + offset, chunkSize);
	return chunkSize;
}

/**
 * Extract ACA header
 */
void ACAFile::LoadHeader(std::istream& ACAStream)
{
	// note that size, CRC and MIME must be replaced after decrypting
	ACAStream.seekg(0);
	ACAStream.rdbuf()->sgetn((char*)m_pDecryptedACA, m_fileList.at(0)->offset);
}

/**
 * Update ACA header
 */

void ACAFile::UpdateHeader(int fileNo, const ARFile& Resource)
{
	m_pLog->LogDebug("Updating fileindex");

	unsigned long hdrPos = m_fileList.at(fileNo)->headerOffset;

	// update offset
	m_pDecryptedACA[hdrPos + 0] = (m_fp & 0xFF000000) >> 24;
	m_pDecryptedACA[hdrPos + 1] = (m_fp & 0x00FF0000) >> 16;
	m_pDecryptedACA[hdrPos + 2] = (m_fp & 0x0000FF00) >> 8;
	m_pDecryptedACA[hdrPos + 3] =  m_fp & 0x000000FF;

	// now it's ok to update our 'virtual' file pointer
	// position it right after the current ARF
	m_fp += Resource.Filesize();

	std::stringstream message;
	message << "fp = " << m_fp;
	m_pLog->LogDebug(message.str());

	// update size
	m_pDecryptedACA[hdrPos + 4] = (Resource.Filesize() & 0xFF000000) >> 24;
	m_pDecryptedACA[hdrPos + 5] = (Resource.Filesize() & 0x00FF0000) >> 16;
	m_pDecryptedACA[hdrPos + 6] = (Resource.Filesize() & 0x0000FF00) >> 8;
	m_pDecryptedACA[hdrPos + 7] =  Resource.Filesize() & 0x000000FF;

	// update CRC
	unsigned long crc32 = 0;
	CRC32 CRC;
	CRC.ArrayCRC32((char*)Resource.ResourceFile(), Resource.Filesize(), crc32);
	m_pDecryptedACA[hdrPos + 8]		= (crc32 & 0xFF000000) >> 24;
	m_pDecryptedACA[hdrPos + 9]		= (crc32 & 0x00FF0000) >> 16;
	m_pDecryptedACA[hdrPos + 10]	= (crc32 & 0x0000FF00) >> 8;
	m_pDecryptedACA[hdrPos + 11]	=  crc32 & 0x000000FF;

	// udate MIME
	ARFile::MIMEType MIME = GetMIME(m_fileList.at(fileNo)->filename);
	m_pDecryptedACA[hdrPos + 12] = MIME;
}

/**
 * Extract file list, including details
 */
void ACAFile::ExtractList(std::istream& ACAStream)
{
	m_pLog->LogInfo("*** Extracting fileindex ***");

	uint8_t ACAHeader[32];
	uint8_t* buffer = new uint8_t[256];
	std::stringstream msg;
	int i;

	// process ACA header
	char	headerID[8];	// HDDVDACA
	uint8_t filesNo[2];		// 0x146 326?
	uint8_t noOfBytes[4];	// 0x013427AD 20.195.245

	// check if we're dealing with a HD DVD ACA file
	/** @todo Check if header is HDDVDACA */

	// make sure we start at the beginning
	ACAStream.seekg(0);
	ACAStream.rdbuf()->sgetn((char*)ACAHeader, sizeof(ACAHeader));

	//
	for (i = 0; i < 2; i++) {
		filesNo[i] = ACAHeader[i + 12];
	}

	m_noOfFiles = filesNo[0] * 0x100 + filesNo[1];
	msg << "No of files: " << m_noOfFiles;
	m_pLog->LogInfo(msg.str());

	for (i = 0; i < 4; i++) {
		noOfBytes[i] = ACAHeader[i + 14];
	}
	m_filesize =  noOfBytes[0] * 0x1000000
				+ noOfBytes[1] * 0x10000
				+ noOfBytes[2] * 0x100
				+ noOfBytes[3];
	msg.str("");
	msg << "Filesize: " << m_filesize << " bytes";
	m_pLog->LogInfo(msg.str());

	// process dataheaders
	int fileNo;
	for (fileNo = 0; fileNo < m_noOfFiles; fileNo++)
	{
		/** @note TELLG() IS LIMITED AND SIGNED!!!! */
		const unsigned long headerOffset = ACAStream.tellg();

		ACAStream.rdbuf()->sgetn((char*)buffer, 4);
		const unsigned long offset = SwitchEndianLong((unsigned long*)buffer);

		// read data size
		ACAStream.rdbuf()->sgetn((char*)buffer, 4);
		const unsigned long dataSize = SwitchEndianLong((unsigned long*)buffer);

		// read CRC checksum
		ACAStream.rdbuf()->sgetn((char*)buffer, 4);
		const unsigned long CRC = SwitchEndianLong((unsigned long*)buffer);

		// read mime (always 0xFF)
		ARFile::MIMEType MIMEValue = (ARFile::MIMEType)ACAStream.rdbuf()->sbumpc();

		// read name length
		int filenameLength = ACAStream.rdbuf()->sbumpc();

		// read filename (variable size!)
		ACAStream.rdbuf()->sgetn((char*)buffer, filenameLength);
		buffer[filenameLength] = '\0';
		const std::string filename = (char*)buffer;

		// read 32 dummy bytes for easy filepointer positioning
		ACAStream.rdbuf()->sgetn((char*)buffer, 32);

		// log details
		msg.str("");
		msg << filename << std::hex
			<< ", hdroffset=0x" << headerOffset
			<< ", offset=0x" 	<< offset
			<< ", size=" 		<< dataSize
			<< ", CRC32=0x" 	<< CRC;
		m_pLog->LogDebug(msg.str());

		// add to filelist
		ResourceFile* acaContent = new ResourceFile;
		acaContent->headerOffset 	= headerOffset;
		acaContent->offset			= offset;
		acaContent->dataSize		= dataSize;
		acaContent->CRC				= CRC;
		acaContent->filename		= filename;
		m_fileList.push_back(acaContent);
	}

	delete[] buffer;
}

/**
 * Dump list of files to console
 */
void ACAFile::DumpFileList()
{
	std::stringstream infoHeader;
	infoHeader << m_noOfFiles << " files, " << m_filesize << " bytes total";

	m_pLog->LogInfo("*** Dumping filelist ***");
	m_pLog->LogInfo(infoHeader.str());
	m_pLog->LogInfo("\n");
	m_pLog->LogInfo("offset\tsize\tCRC\t\ttitle");

	int fileNo;
	for (fileNo = 0; fileNo < m_noOfFiles; fileNo++)
	{
		std::stringstream entry;
		entry	<< m_fileList.at(fileNo)->offset 	<< "\t"
			 	<< m_fileList.at(fileNo)->dataSize	<< "\t"
				<< m_fileList.at(fileNo)->CRC 		<< "\t"
				<< m_fileList.at(fileNo)->filename	<< std::endl;
		m_pLog->LogInfo(entry.str());
	}
}

/**
 * quick and dirty guess MIME-type based on filename
 */
ARFile::MIMEType ACAFile::GetMIME(const std::string& filename)
{
	// get extention
	std::string ext = filename.substr( filename.rfind('.', filename.length()) );
	std::transform(ext.begin(), ext.end(), ext.begin(), (int(*)(int)) tolower);

	if      (ext == ".xpl")	return ARFile::MIME_XPL;
	else if (ext == ".xmf")	return ARFile::MIME_XMF;
	else if (ext == ".xmu")	return ARFile::MIME_XMU;
	else if (ext == ".xts")	return ARFile::MIME_XTS;
	else if (ext == ".xas")	return ARFile::MIME_XAS;
	else if (ext == ".xss")	return ARFile::MIME_XSS;
	else if (ext == ".js")	return ARFile::MIME_JS;
	else if (ext == ".evo")	return ARFile::MIME_EVO;
	else if (ext == ".map")	return ARFile::MIME_MAP;
	else if (ext == ".jpg")	return ARFile::MIME_JPG;
	else if (ext == ".png")	return ARFile::MIME_PNG;
	else if (ext == ".mng")	return ARFile::MIME_MNG;
	else if (ext == ".cvi")	return ARFile::MIME_CVI;
	else if (ext == ".cdw")	return ARFile::MIME_CDW;
	else if (ext == ".wav")	return ARFile::MIME_WAV;
	else if (ext == ".ttf")	return ARFile::MIME_TTF;
	else if (ext == ".xml")	return ARFile::MIME_XML;
	else if (ext == ".xul")	return ARFile::MIME_XUL;
	else if (ext == ".otf")	return ARFile::MIME_OTF;
	else if (ext == ".ttc")	return ARFile::MIME_TTC;
	else
	{
		m_pLog->LogNotice("Unknown MIME-type found: " + ext);
		return ARFile::MIME_UNKNOWN;
	}
}

/**
 * byte[4] to unsigned long
 */
unsigned long ACAFile::SwitchEndianLong(unsigned long* value)
{
	uint8_t* b = (uint8_t*)value;
	return b[0]<<24 | b[1]<<16 | b[2]<<8 | b[3];
}

/**
 * byte[2] to unsigned short
 */
unsigned short ACAFile::SwitchEndianShort(unsigned short* value)
{
	uint8_t* b = (uint8_t*)value;
	return b[0]<<8 | b[1];
}

unsigned long ACAFile::DecryptedFilesize()
{
	return m_decryptedACASize;
}
