/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "arfile.hpp"
#include "rijndael.hpp"
#include "util.hpp"
#include <iostream>
#include <sstream>
#include <stdexcept>

/**
 * Used for decrypting files containing multiple ARFs (eg ACA files)
 */
ARFile::ARFile(Logger* pLog, uint8_t* pTitleKeyTable)
:	m_pLog(pLog),
	m_pTitleKeyTable(pTitleKeyTable),
	m_pRawData(0),
	m_rawDataSize(0x00000000),
	m_isEncrypted(false),
	m_hasHash(false),
	m_hasMAC(false),
	m_protectionType(NON_PROTECTED),
	m_TITLE_KEY_PTR(0x00),
	m_filesize(0x00000000),	// does not include the Resource File Name field of 272 bytes
	m_pResourceFile(0)
{
}

/**
 * When dealing with a single file (TitleKeys not supported yet)
 */
ARFile::ARFile(Logger* pLog, const std::string& filename)
:	m_pLog(pLog),
	m_pTitleKeyTable(0),
	m_pRawData(0),
	m_rawDataSize(0x00000000),
	m_isEncrypted(false),
	m_hasHash(false),
	m_hasMAC(false),
	m_protectionType(NON_PROTECTED),
	m_TITLE_KEY_PTR(0x00),
	m_filesize(0x00000000),	// does not include the Resource File Name field of 272 bytes
	m_pResourceFile(0)
{
	std::ifstream ARFStream(filename.c_str(), std::ios::binary);
	if (!ARFStream.is_open())
	{
		std::stringstream errorMsg;
		errorMsg << "Error opening file " << filename;
		m_pLog->LogError(errorMsg.str());
		throw std::runtime_error(errorMsg.str());
	}

	unsigned long dataOffset = ReadHeader(ARFStream);
	ReadData(ARFStream, dataOffset);
	ExtractResource();
	ARFStream.close();
}

/**
 * Clean up allocated memory
 */
ARFile::~ARFile()
{
	if (m_pRawData)
		delete[] m_pRawData;
	if (m_pResourceFile)
		delete[] m_pResourceFile;
}

/**
 * Read and process ARF header
 *
 * 4 - file ID
 * 1 - protection type
 * 1 - reserved
 * 1 - TITLE_KEY_PTR
 * 4 - file size
 *
 * @param inputStream	Filestream
 * @param offset		Offset in file (default 0)
 * @return No of bytes read or 0 if this wasn't an AACS ARF
 */
unsigned long ARFile::ReadHeader(std::ifstream& inputStream,
								 unsigned long 	offset)
{
	uint8_t IDbuffer[4];

	// read header, starting at offset
	inputStream.seekg(offset);
	inputStream.rdbuf()->sgetn((char*)IDbuffer, sizeof(IDbuffer));

	// check if we're dealing with AACS ARFs
	// should start with "AACS" (0x41 41 43 53)
	std::string fileID = std::string((char*)IDbuffer, 4);
	if (fileID != "AACS")
	//if ((uint32_t)((uint32_t*)IDbuffer) != 0x41414353)
	{
		std::stringstream errorMsg;
		errorMsg << "Invalid AACS ARF at offset 0x" << std::hex << offset << " (" << fileID << ")";
		//errorMsg << "Invalid AACS ARF at offset 0x" << std::hex << offset << " (" << (uint32_t)((uint32_t*)IDbuffer) << ")";
		/** @test Do we want to throw here? Might be usefull to just return 0 without throwing */
		throw std::runtime_error(errorMsg.str());
	}

	// AACS ARF, so continue
	m_protectionType	= (ProtectionType)inputStream.rdbuf()->sbumpc();
	char reserved		= inputStream.rdbuf()->sbumpc();
	m_TITLE_KEY_PTR		= inputStream.rdbuf()->sbumpc();
	m_filesize			= inputStream.rdbuf()->sbumpc() * 0x1000000
						+ inputStream.rdbuf()->sbumpc() * 0x10000
						+ inputStream.rdbuf()->sbumpc() * 0x100
						+ inputStream.rdbuf()->sbumpc();

	std::stringstream message;
	message << "Read ARF header - "
			<< "type="		<< m_protectionType
			<< ", TKptr="	<< m_TITLE_KEY_PTR
			<< ", size=" 	<< m_filesize;
	m_pLog->LogDebug(message.str());

	return 11;
}

/**
 * Read ARF data following header
 *
 * @param inputStream	Input stream to original file
 * @param offset		Offset in original file after header
 */
unsigned long ARFile::ReadData(	std::ifstream& 	inputStream,
								unsigned long 	offset)
{
	// make sure header is read
	if (m_filesize == 0)
		throw std::runtime_error("ARF filesize is 0");

	std::stringstream message;
	message << "Loading raw ARF data (0x" << std::hex << offset << std::dec << ", " << m_filesize << ")";
	m_pLog->LogDebug(message.str());

	// set postion in file
	inputStream.seekg(offset);
	unsigned long readBytes;

	/* encryption */
	if (m_protectionType == ENCRYPTED)
	{
		std::stringstream message;
		message << "Detected encrypted ARF - needs TITLE_KEY_PTR " << (int)m_TITLE_KEY_PTR;
		m_pLog->LogDebug(message.str());

		m_isEncrypted = true;
		m_rawDataSize = 272 + m_filesize;

		try
		{
			/** @todo Prevent for allocating double mem */
			m_pRawData		= new uint8_t[m_rawDataSize];
			m_pResourceFile	= new uint8_t[m_filesize];
		}
		catch (std::bad_alloc& ) {
			throw std::runtime_error("Error allocation memory for ARF");
		}

		// read encrypted data
		inputStream.rdbuf()->sgetn((char*)m_pRawData, m_rawDataSize);
		readBytes = m_rawDataSize;
	}

	/* encryption and hash */
	else if (m_protectionType == ENCRYPTED_HASH)
	{
		std::stringstream message;
		message << "Detected encrypted and hashed ARF - needs TITLE_KEY_PTR " << (int)m_TITLE_KEY_PTR;
		m_pLog->LogDebug(message.str());

		m_isEncrypted	= true;
		m_hasHash		= true;
		m_rawDataSize 	= 272 + m_filesize;		// hashes NOT included!

		try
		{
			m_pRawData		= new uint8_t[m_rawDataSize];
			m_pResourceFile	= new uint8_t[m_filesize];
		}
		catch (std::bad_alloc& ) {
			throw std::runtime_error("Error allocation memory for ARF");
		}

		// read encrypted data
		inputStream.rdbuf()->sgetn((char*)m_pRawData, m_rawDataSize);

		/** @todo hashing stuff (needs CONTENT_HASH_TABLE2.AACS) */
		m_pLog->LogError("Encapsulation Format for Encryption and Hash is NOT supported yet");
		m_pLog->LogError("If you see this, please let me know as I've never seen these before");

		readBytes = m_rawDataSize + 0;	// add hashes size!
	}

	/* MAC */
	else if (m_protectionType == MAC)
	{
		m_pLog->LogDebug("Detected ARF with MAC");
		m_hasMAC		= true;
		m_rawDataSize	= 272 + m_filesize + 16;

		// RFN field
		uint8_t* filename = new uint8_t[272];
		inputStream.rdbuf()->sgetn((char*)filename, 272);
		m_filename = (char*)filename;
		delete[] filename;
		m_pLog->LogDebug(m_filename);

		// read resource data
		try
		{
			m_pResourceFile = new uint8_t[m_filesize];
			inputStream.rdbuf()->sgetn((char*)m_pResourceFile, m_filesize);
		}
		catch (std::bad_alloc& ) {
			throw std::runtime_error("Error allocation memory for ARF");
		}

		// read MAC (16 bytes)
		uint8_t MAC[16];
		inputStream.rdbuf()->sgetn((char*)MAC, sizeof(MAC));
		/** @todo MAC ARF processing */

		readBytes = m_rawDataSize;
	}

	/* hash */
	else if (m_protectionType == HASH)
	{
		m_pLog->LogDebug("Detected hashed ARF");
		m_hasHash		= true;
		m_rawDataSize	= 272 + m_filesize + 4;

		// RFN field
		uint8_t* filename = new uint8_t[272];
		inputStream.rdbuf()->sgetn((char*)filename, 272);
		m_filename = (char*)filename;
		delete[] filename;

		// read resource data
		try
		{
			m_pResourceFile = new uint8_t[m_filesize];
			inputStream.rdbuf()->sgetn((char*)m_pResourceFile, m_filesize);
		}
		catch (std::bad_alloc& ) {
			throw std::runtime_error("Error allocation memory for temp ARF storage");
		}

		// read hash ptr (4 bytes)
		uint8_t hashPointer[4];
		inputStream.rdbuf()->sgetn((char*)hashPointer, sizeof(hashPointer));
		/** @todo Do something useful with hash pointer */

		readBytes = m_rawDataSize;
	}

	/* Non-Protected Advanced Element */
	else if (m_protectionType == NON_PROTECTED)
	{
		m_pLog->LogNotice("Detected Non-Protected Advanced Element ARF");
		m_isEncrypted	= false;	// but name field is encrypted!
		m_rawDataSize 	= 272 + m_filesize;

		// RFN field
		uint8_t* filename = new uint8_t[272];
		inputStream.rdbuf()->sgetn((char*)filename, 272);
		//m_filename = (char*)filename;
		// is encrypted, so can't read it yet
		std::stringstream tmpFilename;
		tmpFilename << "encrypted_filename_" << offset;
		m_filename = tmpFilename.str();
		delete[] filename;

		// read resource data
		try
		{
			m_pResourceFile = new uint8_t[m_filesize];
			inputStream.rdbuf()->sgetn((char*)m_pResourceFile, m_filesize);
		}
		catch (std::bad_alloc& ) {
			throw std::runtime_error("Error allocation memory for temp ARF storage");
		}

		readBytes = m_rawDataSize;
	}

	return readBytes;
}

/**
 * Extract (and decrypt) single ARF
 */
void ARFile::ExtractResource()
{
	m_pLog->LogInfo("Extracting " + m_filename);

	try
	{
		// ARF is encrypted
		// decrypt entire encrypted part (resource file name + content)
		if (m_isEncrypted) {
			Decrypt(m_pTitleKeyTable + m_TITLE_KEY_PTR * 16);
		}

		// ARF has hashes
		if (m_hasHash) {
			m_pLog->LogNotice("Hash check not implemented yet. Assuming file is good");
			/** @todo Verify hash */
		}

		// ARF has a MAC
		if (m_hasMAC) {
			m_pLog->LogNotice("MAC check not implemented yet. Assuming file is good");
			/** @todo Verify MAC */
		}
	}
	catch (std::exception e)
	{
		throw;
	}
}

/**
 * HDDVDFS read data
 *
 * @param offset	Offset in file
 * @param chunkSize	Size of chunck to read
 * @param buffer	Buffer to place read data into
 * @return No of bytes read
 */
unsigned long ARFile::ReadChunk(unsigned long offset,
								unsigned long chunkSize,
								uint8_t* const buffer)
{
	if (offset + chunkSize > m_filesize) {
		m_pLog->LogError("Requested size is bigger than available bytes");
		return 0;
	}

	memcpy(buffer, m_pResourceFile + offset, chunkSize);
	return chunkSize;
}

/**
 * Decrypt ARF
 * Residual data with size less than 16 bytes is left unencrypted
 *
 * @param TitleKeyPointer	Title Key used for decrypting
 */
void ARFile::Decrypt(const uint8_t* const TitleKey)
{
	m_pLog->LogDebug("Decrypting ARF");

	const uint8_t AESCBCConstantIV[] = {
		0x0B, 0xA0, 0xF8, 0xDD, 0xFE, 0xA6, 0x1F, 0xB3,
		0xD8, 0xDF, 0x9F, 0x56, 0x6A, 0x05, 0x0F, 0x78
	};

	// take residual unencrypted bytes into account
	unsigned long encryptedSize = 272 + m_filesize;
	if (encryptedSize % 16 != 0) {
		encryptedSize -= encryptedSize % 16;
	}

	// do actual decryption
	char* pDecryptedData = new char[encryptedSize];
	CRijndael oRijndael;

	// CBC, IV
	oRijndael.MakeKey((char*)TitleKey, (char*)AESCBCConstantIV, 16, 16);
	oRijndael.Decrypt((char*)m_pRawData, pDecryptedData, encryptedSize, CRijndael::CBC);

	// overwrite encrypted data with decrypted data (residual bytes should be left in tact)
	memcpy(m_pRawData, pDecryptedData, encryptedSize);

	// get name
	uint8_t* filename = new uint8_t[272];
	memcpy(filename, m_pRawData, 272);
	m_filename = (char*)filename;

	// get decrypted ARF
	memcpy(m_pResourceFile, m_pRawData + 272, m_filesize);

	delete[] pDecryptedData;
	delete[] filename;
}
