/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "evobfile.hpp"
#include "aacs.hpp"
#include "rijndael.hpp"
#include "util.hpp"
#include <string>
#include <fstream>
#include <iostream>
#include <stdexcept>

/**
 *
 */
EVOBFile::EVOBFile(Logger* pLog, bool writeOutput, bool dumpOutput, unsigned long chunkSize)
:	m_pLog(pLog),
	m_writeOutput(writeOutput),
	m_dumpOutput(dumpOutput),
	m_chunkSize(chunkSize),
	m_TITLE_KEY_PTR(0),
	m_pTitleKeyTable(0)
{
}

/**
 * Decrypt file
 *
 * @param sourceFilename	Path to source file
 * @param destinationPath	Path to destination dir
 * @param pTitleKeyTable	Pointer to Title Key table
 */
void EVOBFile::Decrypt(	const std::string& 		sourceFilename,
						const std::string& 		destinationPath,
						const uint8_t* const	pTitleKeyTable)
{
	// set TK table
	m_pTitleKeyTable = (uint8_t*)pTitleKeyTable;

	// open source
	std::ifstream fis(sourceFilename.c_str(), std::ios::binary);
	if (!fis.is_open())
		throw std::ios_base::failure("Error opening " + sourceFilename);

	// output stream
	std::ofstream fos;
	if (m_writeOutput)
	{
		/** @todo Resume existing file */
		fos.open(destinationPath.c_str(), std::ios::binary);
		if (!fos.is_open()) {
			throw std::ios_base::failure("Error opening " + destinationPath);
		}
	}

	uint8_t 	EVOBpack[EVOBSize];
	uint8_t*	currentChunk 	= new uint8_t[m_chunkSize];
	int			packCounter		= 0;
	int			packsPerChunk	= m_chunkSize / EVOBSize;		// 16.384
	bool 		finish			= false;
	int			i				= 0;

	m_pLog->LogInfo("Starting actual decryption");

	// load initial chunk
	LoadChunk(currentChunk, fis, m_chunkSize);

	// go through entire file
	while (!finish)
	{
		memcpy(EVOBpack, currentChunk + packCounter * EVOBSize, EVOBSize);
		DecryptPack(EVOBpack);
		packCounter++;

		// write to stdout
		if (m_dumpOutput)
		{
			// ugly but functional output to stdout
			FILE* output_stream = freopen(NULL, "wb", stdout);
			fwrite((char*)EVOBpack, sizeof(uint8_t), EVOBSize, output_stream);
		}

		// write to file
		if (m_writeOutput) {
			fos.write((char*)EVOBpack, EVOBSize);
		}

		// load next chuck of data to chew on
		if (packCounter == packsPerChunk)
		{
			packCounter  = 0;
			int bytesRead = LoadChunk(currentChunk, fis, m_chunkSize);

			if (bytesRead == 0) {
				finish = true;
			}
			else {
				if (bytesRead != m_chunkSize) {
					packsPerChunk = bytesRead / EVOBSize;
				}
			}
		}
	}

	if (fis) 	fis.close();
	if (fos)	fos.close();
	delete[] currentChunk;
}

/**
 * HDDVDFS open file
 *
 * Opens the encrypted EVOB file and intiialises stuff needed for decryption
 *
 * @param sourceFilename	EVOB file to open
 * @param pTitleKeyTable	Pointer to Title Key table
 * @param fis				Input stream
 */
void EVOBFile::Open(const std::string& 	sourceFilename,
					const uint8_t* const pTitleKeyTable,
					std::ifstream&		fis)
{
	// set TK table
	m_pTitleKeyTable = (uint8_t*)pTitleKeyTable;

	// input stream
	fis.open(sourceFilename.c_str(), std::ios::binary);
	if (!fis.is_open()) {
		throw std::ios_base::failure("Error opening " + sourceFilename);
	}

	// Need to read header to initialize some private values before
	// allowing random reads
	uint8_t	currentChunk[m_chunkSize];
	ReadChunk(fis, currentChunk);
	fis.seekg(0);
}

/**
 * HDDVDFS read EVOB data
 *
 * @param fis
 * @param currentChunk	Buffer containing encrypted data. Will be overwritten by
 * the decrypted packs
 * @return No of bytes read
 */
int EVOBFile::ReadChunk(std::ifstream& fis, uint8_t* const currentChunk)
{
	uint8_t EVOBpack[EVOBSize];
	int packCounter	= 0;
	int bytesRead 	= LoadChunk(currentChunk, fis, m_chunkSize);

	while (packCounter < (bytesRead / EVOBSize))
	{
		DecryptPack(currentChunk + packCounter * EVOBSize);
		packCounter++;
	}

	return bytesRead;
}

/**
 * Load chunk into buffer
 *
 * @param buffer		Buffer to place read data into
 * @param inputStream	Input stream
 * @param blockLen		Size of chunk to read
 * @return No of bytes read
 */
int EVOBFile::LoadChunk(uint8_t* const buffer, std::ifstream &inputStream, int blockLen)
{
	std::streamsize bytesRead = 0;				// no of bytes read

	// read chunk
	while (bytesRead != blockLen)
	{
		inputStream.read((char*)buffer+bytesRead, blockLen-bytesRead);

		// check if EOF is reached
		if (inputStream.eof()) {
			inputStream.clear();
			break;
		}

		bytesRead += inputStream.gcount();
	}

	return bytesRead;
}

/**
 * Pack detection and decryption
 * @param rawData	Raw EVOB pack to decrypt
 */
void EVOBFile::DecryptPack(uint8_t* const rawData)
{
	int i = 0;

	/* detect pack */
	// start of PES pack
	uint32_t PESHeader =	(rawData[0] << 24) | (rawData[1] << 16) |
							(rawData[2] << 8)  |  rawData[3];

	if (PESHeader == PACK_START_CODE)
	{
		uint32_t packPos	= 0x0E;										// skip pack header (prefix + 10 = 0x0E)
		PESHeader			=	(rawData[packPos] << 24)    | (rawData[packPos + 1] << 16) |
								(rawData[packPos + 2] << 8) |  rawData[packPos + 3];

		// system header (= NAV_PCK)
		if (PESHeader == SYSTEM_HEADER_START_CODE)
		{
			packPos				+= 4;												// skip prefix and ID
			uint16_t packLength  = (rawData[packPos] << 8) | rawData[packPos + 1];	// 2 byte length
			packPos				+= 2 + packLength;									// skip length

			// go to next pack
			PESHeader	=	(rawData[packPos] << 24)    | (rawData[packPos + 1] << 16) |
							(rawData[packPos + 2] << 8) |  rawData[packPos + 3];

			// check if Title Key Pointer is valid
			// bits 7:6 => 10b: The Title Key Pointer is valid
			// extract CPI field
			m_CPIField[0]  = rawData[0x3C];
			m_CPIField[1]  = rawData[0x3D];
			m_CPIField[2]  = rawData[0x3E];
			m_CPIField[3]  = rawData[0x3F];
			m_CPIField[4]  = rawData[0x40];
			m_CPIField[5]  = rawData[0x41];
			m_CPIField[6]  = rawData[0x42];
			m_CPIField[7]  = rawData[0x43];
			m_CPIField[8]  = rawData[0x44];
			m_CPIField[9]  = rawData[0x45];
			m_CPIField[10] = rawData[0x46];
			m_CPIField[11] = rawData[0x47];
			m_CPIField[12] = rawData[0x48];
			m_CPIField[13] = rawData[0x49];
			m_CPIField[14] = rawData[0x4A];
			m_CPIField[15] = rawData[0x4B];

			// get TITLE_KEY_PTR (range 0 to 63)
			if ((m_CPIField[0] & 0x80) != 0) 			// check KEY_VF
			{
				m_TITLE_KEY_PTR = m_CPIField[2] - 1;	// note: keylist starts at 1, not 0

				// update Title Key
				for (i = 0; i < 16; i++) {
					m_TitleKey[i] = *(m_pTitleKeyTable + m_TITLE_KEY_PTR * 16 + i);
				}
			}

			// remove copy protection
			rawData[0x3C] = 0x00;		// KEY_VF is no longer valid
			rawData[0x48] = 0x00;		// reset CCI field (copy freely, analog protect off, HD anlog output allowed)
		}

		// private stream 2 (ADV_PKT and ADV_PKT with padding)
		// ID 0xBF at loc 0x0E, not valid otherwise
		else if (PESHeader == PRIVATE_STREAM_2) {
			//std::cout << "found ADV_PKT (unsupported!)" << std::endl;
		}

		// video/audio/whatever data
		else
		{
			// check if pack is encrypted (bit 4 and 5)
			// PES_scrambling_control
			if ((rawData[0x14] & 0x30) == 0x10)
			{
				// extract Title Key Data's first 4 bytes
				// and then Title Key Data || CPI_lsb96
				m_TitleKeyData[0]  = rawData[0x54];
				m_TitleKeyData[1]  = rawData[0x55];
				m_TitleKeyData[2]  = rawData[0x56];
				m_TitleKeyData[3]  = rawData[0x57];
				m_TitleKeyData[4]  = m_CPIField[4];
				m_TitleKeyData[5]  = m_CPIField[5];
				m_TitleKeyData[6]  = m_CPIField[6];
				m_TitleKeyData[7]  = m_CPIField[7];
				m_TitleKeyData[8]  = m_CPIField[8];
				m_TitleKeyData[9]  = m_CPIField[9];
				m_TitleKeyData[10] = m_CPIField[10];
				m_TitleKeyData[11] = m_CPIField[11];
				m_TitleKeyData[12] = m_CPIField[12];
				m_TitleKeyData[13] = m_CPIField[13];
				m_TitleKeyData[14] = m_CPIField[14];
				m_TitleKeyData[15] = m_CPIField[15];

				// decrypt content key (AES ECB)
				uint8_t ContentKey[TK_LENGTH];
				CRijndael oRijndael;
				oRijndael.MakeKey((char*)m_TitleKey, CRijndael::sm_chain0, TK_LENGTH, TK_LENGTH);
				oRijndael.DecryptBlock((char*)m_TitleKeyData, (char*)ContentKey);

				// XOR content key
				for (i = 0; i < 16; i++) {
					ContentKey[i] ^= m_TitleKeyData[i];
				}

				// decrypt video content (AES CBC, using IV)
				uint8_t decryptedData[EVOBContentSize];
				oRijndael.ResetChain();
				oRijndael.MakeKey((char*)ContentKey, (char*)AESCBCConstantIV, TK_LENGTH, TK_LENGTH);
				oRijndael.Decrypt((char*)rawData + EVOBHeaderSize, (char*)decryptedData, EVOBContentSize, CRijndael::CBC);

				// overwrite original encrypted data with decrypted data
				memcpy(rawData + EVOBHeaderSize, decryptedData, EVOBContentSize);

				// remove PES from header
				// Pack encrypted => PES_scrambling_control is 01b
				// not encrypted  => PES_scrambling_control is 00b (this is what we want)
				rawData[0x14] &= 0xCF;		// bits 5:4
			}
		}
	}
}
