/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "filesystem.hpp"
#include "file.hpp"
#include <vector>
#include <algorithm>	// transform
#include <cctype>		// tolower
#include <iostream>
#include <sstream>
#include <stdexcept>

#if defined(_MSC_VER) || defined(__WIN32__)
//#include <dir.h>
#else
#include <sys/dir.h>
#endif
#if defined(_MSC_VER)
#include <winx/direntx.h>
#else
#include <dirent.h>
#endif
#include <errno.h>
#include <sys/stat.h>

/**
 * does nothing atm
 */
Filesystem::Filesystem()
{
}

/**
 * Scan a directory
 * needs seperator at the end (not checked!)
 * 64 bit functions do not compile on windows using mingw
 */
void Filesystem::ScanDir(const std::string& directoryPath, bool recursive)
{
	DIR *dir = opendir( directoryPath.c_str() );

	if (NULL != dir)
	{
		struct dirent64 *entry;

		for(; NULL != (entry = readdir64(dir)); )
		{
			struct stat64 st;

			// Skip dots
			if ('.' == entry->d_name[0] && ( '\0' == entry->d_name[1] ||
				( '.' == entry->d_name[1] && '\0' == entry->d_name[2])) )
			{
				// do nothing
			}
			else
			{
				if (0 == stat64((std::string(directoryPath) + entry->d_name).c_str(), &st))
				{

					// directories
					if (S_IFDIR == (st.st_mode & S_IFDIR))
					{
						// add to FileList
						File *listEntry = new File(directoryPath, "", 0, true);
						m_fileList.push_back(listEntry);

						// when recursive scanning
						if (recursive) {
							std::string newPath = directoryPath + entry->d_name + SEPERATOR;
							ScanDir( newPath );
						}
					}

					// files
					if (S_IFREG == (st.st_mode & S_IFREG))
					{
						std::string fileName = entry->d_name;

						// check for extention
						if (m_extention.empty())
						{
							File *listEntry = new File(directoryPath, fileName);
							m_fileList.push_back(listEntry);
						}
						else
						{
							std::string ext = fileName.substr( fileName.rfind('.', fileName.length()) );
							transform(ext.begin(), ext.end(), ext.begin(), (int(*)(int)) tolower);

							if (ext == m_extention)
							{
								File *listEntry = new File(directoryPath, fileName);
								m_fileList.push_back(listEntry);
							}
						}
					}		// files
				}		// stat
			}		// else
		}		// for

		closedir( dir );
	}
}

/**
 * Create directory
 *
 * @note Does not work with symbolic links
 */
void Filesystem::CreateDir(const std::string& dirName, bool deleteExisting)
{
	int result(-1);
	errno = 0;

#ifdef __WIN32__
	result = mkdir(dirName.c_str());
#else
	result = mkdir(dirName.c_str(), S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#endif

	// error handling
	// there are many more errors not listed here
	if (result != 0)
	{
		switch (errno)
		{
			case EEXIST:
				if (deleteExisting) {
					if (rmdir(dirName.c_str()) != 0)
						throw std::runtime_error("Error removing existing directory (must be empty)");
					CreateDir(dirName, deleteExisting);
				}
				break;

			case EACCES:
				throw std::runtime_error("Error creating directory: Permission denied");
				break;

			case EINVAL:
				throw std::runtime_error("Error creating directory: Invalid argument");
				break;

			default:
				std::stringstream errorMsg;
				errorMsg << "Error creating directory " << dirName << " (error " << errno << ")";
				throw std::runtime_error(errorMsg.str());
		}
	}
}

/**
 * Remove/delete directory
 * dir must be empty or rmdir() fails!
 */
void Filesystem::RemoveDir(const std::string& dirName)
{
	errno = 0;

	if (rmdir(dirName.c_str()) != 0) {
		std::stringstream errorMsg;
		errorMsg << "Error removing directory " << dirName << " (error " << errno << ")";
		throw std::runtime_error(errorMsg.str());
	}
}

