/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "hddrive.hpp"
#include "aacs_crypto.hpp"
#include "backup_hostdata.hpp"
#include "util.hpp"
#include <iostream>
#include <sstream>
#include <stdexcept>

/**
 * Setup HD DVD drive and open device
 */
HDDrive::HDDrive(Logger* pLog, const std::string& device, AACSCrypto* pCrypto)
:	m_pLog(pLog),
	m_hHandle(0),
	m_isAuthenticated(false),
	m_pCrypto(pCrypto)
{
	// make sure all buffers are clean
	memset(m_DriveNonce, 		0x00, sizeof(m_DriveNonce));
	memset(m_DriveCertificate, 	0x00, sizeof(m_DriveCertificate));
	memset(m_DriveKeyPoint, 	0x00, sizeof(m_DriveKeyPoint));
	memset(m_DriveKeySignature,	0x00, sizeof(m_DriveKeySignature));
	memset(m_HostKeySignature, 	0x00, sizeof(m_HostKeySignature));
	memset(m_BusKey, 			0x00, sizeof(m_BusKey));
	memset(m_VolumeID, 			0x00, sizeof(m_VolumeID));
	memset(m_VolumeIDMAC, 		0x00, sizeof(m_VolumeIDMAC));

	// load host keys
	ReadHostData();

	// open drive
	m_hHandle = open_drive((char*)device.c_str());

	if (m_hHandle < 0) {
		throw std::runtime_error("Error opening device " + device);
	}

	std::stringstream ssMsg;
	ssMsg << "Device " << device << " opened (handle " << m_hHandle << ")";
	m_pLog->LogInfo(ssMsg.str());
}

/**
 *
 */
HDDrive::~HDDrive()
{
	if (m_hHandle)
		close_drive(m_hHandle);
}

/**
 * Do drive authentication
 */
void HDDrive::Authenticate()
{
	if (m_isAuthenticated) {
		return;
	}

	// get drive info
	m_pLog->LogInfo("Getting drive info");
	if (!GetDriveInfo()) {
		throw std::runtime_error("Error getting drive info");
	}

	m_pLog->LogDebug("\tVendor\t\t" + m_vendorID);
	m_pLog->LogDebug("\tProduct\t\t" + m_productID);
	m_pLog->LogDebug("\tRevision\t" + m_revision);
	m_pLog->LogDebug("\tDate\t\t" + m_date);
	m_pLog->LogDebug("Starting drive authentication");

	// make sure all 4 drive auth threads are cleared
	InvalidateAGID(0);
	InvalidateAGID(1);
	InvalidateAGID(2);
	InvalidateAGID(3);

	/*
	5. The host acquires an Authentication Grant Identifier (AGID).
	The AGID is used to manage multiple threads of drive authentications up
	to 4, if the drive supports the multiple threads.
	*/
	int8_t AGID = ReportAGID();

	if (AGID < 0) {
		throw std::runtime_error("Error getting AGID");
	}

	// check if we can use the hack
	if (m_productID == "DVD/HD  X807616 " && m_revision == "MC08")
	{
		m_pLog->LogInfo("\tFound Xbox360 HD DVD drive supporting authentication hack");
		ReadVolumeID_hack(0, m_VolumeID, m_VolumeIDMAC);
		m_isAuthenticated = true;
		return;
	}

	//unknown revision, so start real authentication
	AACSAuthentication(AGID);
	VerifyVolumeID();
}

/**
 * Verify VOlume ID using CMAC
 */
void HDDrive::VerifyVolumeID()
{
	m_pLog->LogInfo("Verifying Volume ID");
	m_pLog->LogDebug("\tVolume ID\t" + Utils::ToHexString(m_VolumeID, 16));
	m_pLog->LogDebug("\tVolume ID MAC\t" + Utils::ToHexString(m_VolumeIDMAC, 16));

	if (!m_pCrypto->VolumeIDIsValid(m_BusKey, m_VolumeID, m_VolumeIDMAC)) {
		throw std::runtime_error("Volume ID verification failed!");
	}

	m_pLog->LogInfo("\tVolume ID is valid");
	m_isAuthenticated = true;
}

/**
 * Verify Volume ID using CMAC
 */
bool HDDrive::IsVolumeIDValid()
{
	if (!m_pCrypto->VolumeIDIsValid(m_BusKey, m_VolumeID, m_VolumeIDMAC)) {
		return false;
	}

	m_isAuthenticated = true;
	return true;
}

/**
 * Read host data from disk (certificate, keys, etc)
 *
 * Check if data is available externaly. If not, use internal data (can be outdated)
 */
void HDDrive::ReadHostData()
{
	m_pLog->LogInfo("Reading host certificate, keys, etc.");

	int noOfFilesPresent = ReadHostDataFromDisk();

	// check if all data is accounted for
	if (noOfFilesPresent != 5)
	{
		m_pLog->LogNotice("Not all host data was read. Using internal host data");
		memcpy(m_HostNonce,			HostNonce, 		 20);
		memcpy(m_HostCertificate,	HostCertificate, 92);
		memcpy(m_HostKeyPoint,		HostKeyPoint, 	 40);
		memcpy(m_HostKey, 			HostKey, 		 20);
		m_HostPrivateKey = HostPrivateKey;
	}
}

/**
 * Read host data from external files
 *
 * @return No of external files read
 */
int HDDrive::ReadHostDataFromDisk()
{
	int noOfFilesPresent = 5;	// files needed tomake a complete set

	// read host nonce
	std::ifstream ifNonce("keys/host_nonce.hex", std::ios::binary);
	if (!ifNonce.is_open()) {
		m_pLog->LogError("Error loading host nonce");
		noOfFilesPresent--;
	}
	ifNonce.rdbuf()->sgetn((char*)m_HostNonce, 20);
	ifNonce.close();

	// read host certificate
	std::ifstream ifCertificate("keys/host_certificate.hex", std::ios::binary);
	if (!ifCertificate.is_open()) {
		m_pLog->LogError("Error loading host certificate");
		noOfFilesPresent--;
	}
	ifCertificate.rdbuf()->sgetn((char*)m_HostCertificate, 92);
	ifCertificate.close();

	// read host key
	std::ifstream ifKey("keys/host_key.hex", std::ios::binary);
	if (!ifKey.is_open()) {
		m_pLog->LogError("Error loading host key");
		noOfFilesPresent--;
	}
	ifKey.rdbuf()->sgetn((char*)m_HostKey, 20);
	ifKey.close();

	// read host keypoint
	std::ifstream ifKeyPoint("keys/host_keypoint.hex", std::ios::binary);
	if (!ifKeyPoint.is_open()) {
		m_pLog->LogError("Error loading host keypoint");
		noOfFilesPresent--;
	}
	ifKeyPoint.rdbuf()->sgetn((char*)m_HostKeyPoint, 40);
	ifKeyPoint.close();

	// read host private key (text)
	std::ifstream ifPrivateKey("keys/host_private_key.hex");
	if (!ifPrivateKey.is_open()) {
		m_pLog->LogError("Error loading host private key");
		noOfFilesPresent--;
	}
	ifPrivateKey >> m_HostPrivateKey;

	return noOfFilesPresent;
}

/**
 * Do AACS authentication with drive
 */
bool HDDrive::AACSAuthentication(uint8_t AGID)
{
	/*
	6. The host generates 160 bits random number as nonce Hn
	7. The host sends the nonce Hn generated in step 6 and the Host Certificate to the drive.
	*/
	SendHostCertificate(AGID, m_HostNonce, m_HostCertificate);

	/*
	8. The drive checks if the value of the Certificate Type is 0216 and the value of the Length is 005C16 in the Host Certificate.
	9. The drive verifies the signature of the Host Certificate using the AACS LA Public Key.
	10. The drive checks the Host Revocation List determined in either step 3 or step 4 to ensure that the Host ID of the Host Certificate has not been revoked.
	11. The host sends a request to the drive to return a nonce Dn and the Drive Certificate.
	12. The drive generates 160 bits random number as nonce Dn
	13. The drive sends the nonce Dn generated in step 11 and the Drive Certificate to the host.
	*/
	ReportDriveCertificate(AGID, m_DriveNonce, m_DriveCertificate);

	/*
	14. The host checks if the value of the Certificate Type is 0116 and the value of the Length is 005C16 in the Drive Certificate.
	15. The host verifies the signature of the Drive Certificate using the AACS LA Public Key.

		AACS_Verify(AACS_LApub, Drive_Certsig, Drive_Cert)

		where Drive_Certsig is Byte 52 through Byte 91 of the Drive Certificate and Drive_Cert is Byte 0 through Byte 51 of the Drive Certificate.
	16. The host checks the Drive Revocation List determined in either step 1 or step 2 to ensure that the Drive ID of the Drive Certificate has not been revoked.
	17. The host sends a request to the drive to return a point on the elliptic curve Dv and its associated signature.
	18. The drive generates 160 bits random number as Dk
	19. The drive calculates a point on the elliptic curve Dv.
	20. The drive creates a digital signature of the concatenation of the nonce Hn received in step 7 and the point on the elliptic curve Dv calculated in step 17.
	21. The drive sends the point on the elliptic curve Dv calculated in step 17 and the digital signature Dsig created in step 18 to the host.
	*/
	ReportDriveKey(AGID, m_DriveKeyPoint, m_DriveKeySignature);

	/*
	22. The host verifies the signature of the concatenation of the nonce Hn and the point on the elliptic curve Dv.

		AACS_Verify(AACS_Drivepub, Dsig , Hn || Dv)

	23. The host generates 160 bits random number as Hk
	24. The host calculates a point on the elliptic curve Hv.

		Hv = Hk G
		where G is the base point of the elliptic curve
	25. The host creates a digital signature of the concatenation of the nonce Dn received in step 12 and the point on the elliptic curve Hv calculated in step 22.

		Hsig = AACS_Sign(AACS_Hostpriv, Dn || Hv)
	*/
	if (!m_pCrypto->AACS_Sign(m_HostCertificate, m_HostPrivateKey,
							  m_HostKeySignature, m_DriveNonce, m_HostKeyPoint)) {
		throw std::runtime_error("AACS_Sign failed");
	}

	// verify drive signature
	if (!m_pCrypto->AACS_Verify(m_DriveCertificate, m_DriveKeySignature,
								m_HostNonce, m_DriveKeyPoint)) {
		throw std::runtime_error("Drive signature verification failed");
	}

	// verify host signature
	if (m_pCrypto->AACS_Verify(m_HostCertificate, m_HostKeySignature,
							   m_DriveNonce, m_HostKeyPoint) <= 0) {
		throw std::runtime_error("Host signature verification failed");
	}

	/*
	26. The host sends the point on the elliptic curve Hv calculated in step 22 and the digital signature Hsig created in step 23 to the drive.
	27. The drive verifies the signature of the concatenation of the nonce Dn and the point on the elliptic curve Hv.
	*/
	SendHostKey(AGID, m_HostKeyPoint, m_HostKeySignature);

	/*
	28. The drive calculates the Bus Key from a point of the elliptic curve
	29. The host calculates the Bus Key from a point of the elliptic curve
	*/
	m_pCrypto->CalculateBusKey(m_HostCertificate, m_HostKey, m_DriveKeyPoint, m_BusKey);

	/*
	1. The drive and the host carry out the drive authentication and key sharing (AACS-Auth)
	2. The host sends a request to the drive to return Volume ID from the drive
	3. The drive reads Volume ID  and calculates a message authentication code (Dm) from the Volume ID and the Bus Key calculated in step 26
	4. The drive sends Volume_ID read in step 3 and the message authentication code Dm calculated in step 3
	*/
	ReadVolumeID(AGID, m_VolumeID, m_VolumeIDMAC);

	/*
	6. The host verifies if the Dm received in step 4 matches the Hm calculated in step 5. If the verification succeeds, then the host may trust the Volume ID; otherwise, the host should stop processing the media.
	*/

	return true;
}

/**
 * Invalidate AGID
 *
 * Reset drive and clears old challenge stuff?
 *
 * @param AGID Authentication Grant Identifier
 * @return True if successfull
 */
bool HDDrive::InvalidateAGID(uint8_t AGID)
{
	uint8_t buf[1];

	if (!ReportKey(buf, AGID, 63, 0)) {
		throw std::runtime_error("Error invalidating AGID");
	}

	return true;
}

/**
 * Get Authentication Grant Identifier
 */
int8_t HDDrive::ReportAGID()
{
	uint8_t buf[8];
	memset(buf, 0x00, sizeof(buf));

	if (!ReportKey(buf, 0, 0, 8)) {
		throw std::runtime_error("Error getting AGID");
	}

	// check REPORT KEY data length
	if (buf[0] != 0 || buf[1] != 0x06) {
		return -2;
	}

	// return AGID (byte 7, bit 6 and 7)
	return (buf[7] & 0xFF) >> 6;
}

/**
 * Get Drive Nonce and Certificate
 *
 * Certificate contains drive ID, public key and signature
 *
 * @param AGID			Authentication Grant Identifier
 * @param Nonce			Resulting Drive Nonce
 * @param Certificate	Resulting Drive Certificate
 * @return True if successful
 */
bool HDDrive::ReportDriveCertificate(uint8_t AGID, uint8_t* Nonce, uint8_t* Certificate)
{
	uint8_t buf[116];
	memset(buf, 0x00, sizeof(buf));

	if (!ReportKey(buf, AGID, 1, 116)) {
		throw std::runtime_error("Error fetching drive certificate");
	}

	// check REPORT KEY data length
	if (buf[0] != 0 || buf[1] != 0x72) {
		throw std::runtime_error("Error fetching drive certificate - invalid length");
	}

	// place result in Nonce and Certificate
	memcpy(Nonce, buf + 4, 20);				// nonce, 20 bytes, offset 4
	memcpy(Certificate, buf + 24, 92);		// certificate, 92 bytes, offset 24

	return true;
}

/**
 * Get Drive Point and Signature
 *
 * @param AGID		Authentication Grant Identifier
 * @param Point		Resulting Drive Point
 * @param Signature	Resulting Drive Signature
 * @return True if successful
 */
bool HDDrive::ReportDriveKey(uint8_t AGID, uint8_t* Point, uint8_t* Signature)
{
	uint8_t buf[84];
	memset(buf, 0x00, sizeof(buf));

	if (!ReportKey(buf, AGID, 2, 84)) {
		throw std::runtime_error("Error fetching drive key");
	}

	// check REPORT KEY data length
	if (buf[0] != 0 || buf[1] != 0x52) {
		throw std::runtime_error("Error fetching drive key - invalid length");
	}

	// place result in Point and Signature
	memcpy(Point, buf + 4, 40);				// point, 40 bytes, offset 4
	memcpy(Signature, buf + 44, 40);		// signature, 40 bytes, offset 44

	return true;
}

/**
 * Send Host Certificate to drive
 *
 * @param AGID			Authentication Grant Identifier
 * @param Nonce			Host Nonce
 * @param Certificate	Host Certificate
 * @return True if successful
 */
bool HDDrive::SendHostCertificate(uint8_t AGID, uint8_t* Nonce, uint8_t* Certificate)
{
	uint8_t buf[116];
	memset(buf, 0x00, sizeof(buf));

	// SEND KEY parameter list length
	buf[0] = 0x00;
	buf[1] = 0x72;

	// place Nonce and Certificate in buffer
	memcpy(buf + 4, Nonce, 20);				// length 20, offset 4
	memcpy(buf + 24, Certificate, 92);		// length 92, offset, 24

	return SendKey(buf, AGID, 1, 116);
}

/**
 * Send Host Key to drive
 *
 * @param AGID		Authentication Grant Identifier
 * @param point		Host Point
 * @param signature	Host Signature
 * @return True if successful
 */
bool HDDrive::SendHostKey(uint8_t AGID, uint8_t* Point, uint8_t* Signature)
{
	uint8_t buf[84];
	memset(buf, 0x00, sizeof(buf));

	// SEND KEY data length
	buf[0] = 0x00;
	buf[1] = 0x52;

	// place point and signature in buffer
	memcpy(buf + 4, Point, 40);			// length 40, offset 4
	memcpy(buf + 44, Signature, 40);	// length 40, offset 44

	return SendKey(buf, AGID, 2, 84);
}

/**
 * Read Volume ID
 * Is a READ DISK STRUCTURE command extention
 *
 * @param AGID		Authentication Grant Identifier
 * @param pVolumeID	Resulting Volume ID
 * @param pMAC		Resulting MAC
 * @return True if successful
 */
bool HDDrive::ReadVolumeID(uint8_t AGID, uint8_t* VolumeID, uint8_t* MAC)
{
	uint8_t buf[36];
	ReadDiskStructure(buf, AGID, 36);

	// extract Volume ID and MAC
	memcpy(VolumeID, buf + 4, 16);		// length 16, offset 4
	memcpy(MAC, buf + 20, 16);			// length 16, offset 20

	return true;
}

/**
 * Read Volume ID using hack
 */
bool HDDrive::ReadVolumeID_hack(uint8_t AGID, uint8_t* VolumeID, uint8_t* MAC)
{
	// enable DF command
	if (!EnableCommandDF()) {
		m_pLog->LogError("Error enabling DF command");
		return false;
	}

	// write 0001 to 60C1CE and 0005 to 60C1CA
	const uint32_t offset = 0x0060C1C8;
	CommandDfMemset(offset + 4, 2, 0x0000);
	CommandDfMemset(offset + 6, 2, 0x0001);
	CommandDfMemset(offset    , 2, 0x0000);
	CommandDfMemset(offset + 2, 2, 0x0005);

	// extract Volume ID
	uint8_t buf[36];
	if (!ReadDiskStructure(buf, AGID, 36))
	{
		m_pLog->LogError("Error reading Volume ID");
		DisableCommandDF();
		return false;
	}

	// extract Volume ID and MAC
	memcpy(VolumeID, buf + 4, 16);		// length 16, offset 4
	memcpy(MAC, buf + 20, 16);			// length 16, offset 20

	// disable DF command
	if (!DisableCommandDF()) {
		m_pLog->LogError("Error disabling DF command");
		return false;
	}

	return true;
}

/**
 *
 */
bool HDDrive::CommandDfMemset(uint32_t offset, size_t size, uint16_t value)
{
	const uint32_t end = offset + size - 1;

	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);
	m_cmd[0] = 0xDF;
	m_cmd[2] = 0xE3;
	m_cmd[4]  = (offset >> 16)	& 0xFF;
	m_cmd[5]  = (offset >> 8) 	& 0xFF;
	m_cmd[6]  =  offset 		& 0xFF;
	m_cmd[7]  = (end >> 16) 	& 0xFF;
	m_cmd[8]  = (end >> 8) 		& 0xFF;
	m_cmd[9]  =  end 			& 0xFF;
	m_cmd[10] = (value >> 8) 	& 0xFF;
	m_cmd[11] =  value 			& 0xFF;

	return (send_cmd(m_hHandle, m_cmd, 0, 0, 0) == 0);
}

/**
 * Enable DF command
 */
bool HDDrive::EnableCommandDF()
{
	uint8_t buf[8];
	memset(buf, 0x00, sizeof(buf));

	// subcmd
	buf[0] = 0x88;	//
	buf[1] = 0x00;	//
	buf[2] = 0x00;	//
	buf[3] = 0x04;	//
	buf[4] = 0x02;	//
	buf[5] = 0x6F;	//
	buf[6] = 0x01;	//
	buf[7] = 0x00;	//

	return SendDiagnostic(buf, 8);
}

/**
 * Disable DF command
 */
bool HDDrive::DisableCommandDF()
{
	uint8_t buf[8];
	memset(buf, 0x00, sizeof(buf));

	buf[0] = 0x88;	//
	buf[1] = 0x00;	//
	buf[2] = 0x00;	//
	buf[3] = 0x04;	//
	buf[4] = 0x02;	//
	buf[5] = 0x6F;	//
	buf[6] = 0x00;	//
	buf[7] = 0x00;	//

	return SendDiagnostic(buf, 8);
}

/**
 * Get firmware information (not supported?! feature 0x010C is not present)
 *
 * @param firmware Resulting firmware info (timestamp format: CCYYMMDDhhmmss)
 * @return True if successful
 */
/*
bool HDDrive::GetFirmware(uint8_t* firmware)
{
	uint8_t buf[24];
	GetConfiguration(buf, 0x010C, 24);
	memcpy(firmware, buf + 8, 14);		// length 14, offset 8
}
*/

/**
 *
 */
bool HDDrive::GetDriveInfo()
{
	uint8_t buf[96];

	if (!Inquiry(buf, 96))
		return false;

	std::string inquiryInfo((char*)buf, 96);
	m_vendorID	= inquiryInfo.substr(8, 8);
	m_productID	= inquiryInfo.substr(16, 16);
	m_revision	= inquiryInfo.substr(32, 4);
	m_date		= inquiryInfo.substr(36, 8);

	return true;
}

/**
 * Get key from drive
 *
 * @param buffer	Buffer where key will be stored
 * @param AGID		Authentication Grant Identifier
 * @param KeyFormat	See table 4.6 for Key Format Code Definition
 * @param length	Allocation length
 * @return True if successful
 */
bool HDDrive::ReportKey(uint8_t* buffer, uint8_t AGID, uint8_t KeyFormat, short length)
{
	memset(buffer, 0x00, length);
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);

	m_cmd[0] = 0xA4;							// operation code
	m_cmd[7] = 0x02;							// Key Class
	m_cmd[8] = (length >> 8) & 0xFF;			// allocation length (MSB)
	m_cmd[9] = length & 0xFF;					// allocation length (LSB)
	m_cmd[10] = AGID << 6 | (KeyFormat & 0x3F);	// AGID and Key Format

	// send to drive
	return (send_cmd(m_hHandle, m_cmd, buffer, 0, length) == 0);
}

/**
 * Send key to drive
 *
 * @param buffer	Buffer containing data to send
 * @param AGID		Authentication Grant Identifier
 * @param KeyFormat	See table 4.6 for Key Format Code Definition
 * @param length	Allocation length
 * @return True if successful
 */
bool HDDrive::SendKey(uint8_t* pBuffer, uint8_t AGID, uint8_t KeyFormat, short length)
{
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);

	m_cmd[0] = 0xA3;							// operation code
	m_cmd[7] = 0x02;							// Key Class
	m_cmd[8] = (length >> 8) & 0xFF;			// parameter list length, MSB
	m_cmd[9] = length & 0xFF;					// parameter list length, LSB
	m_cmd[10] = AGID << 6 | (KeyFormat & 0x3F);	// AGID and Key Format

	// send to drive
	return send_cmd(m_hHandle, m_cmd, pBuffer, length, 0);
}

/**
 * Get disk structure info from drive
 *
 * @param buffer	Buffer where read data will be stored
 * @param AGID		Authentication Grant Identifier
 * @param length	Allocation length
 * @return True if successful
 */
bool HDDrive::ReadDiskStructure(uint8_t* buffer, uint8_t AGID, short length)
{
	memset(buffer, 0x00, length);
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);

	m_cmd[0] = 0xAD;					// operation code
	m_cmd[1] = 0x00;					// media type
	m_cmd[7] = 0x80;					// format code
	m_cmd[8] = (length >> 8) & 0xFF;	// allocation length (MSB)
	m_cmd[9] = length & 0xFF;			// allocation length (LSB)
	m_cmd[10] = (AGID << 6) & 0xC0;		// AGID

	// send to drive
	return (send_cmd(m_hHandle, m_cmd, buffer, 0, length) == 0);
}

/**
 * GET CONFIGURATION command
 */
bool HDDrive::GetConfiguration(uint8_t* buffer, uint16_t feature, short length)
{
	memset(buffer, 0x00, length);
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);

	// CDB
	m_cmd[0] = 0x46;					// operation code
	m_cmd[1] = 0x00;					// RT
	m_cmd[2] = (feature >> 8) & 0xFF;	// starting feature number (MSB)
	m_cmd[3] = feature & 0xFF;			// starting feature number (LSB)
	m_cmd[7] = (length >> 8) & 0xFF;	// allocation length (MSB)
	m_cmd[8] = length & 0xFF;			// allocation length (LSB)

	// send to drive
	return (send_cmd(m_hHandle, m_cmd, buffer, 0, length) == 0);
}

/**
 * INQUIRY command
 */
bool HDDrive::Inquiry(uint8_t* buffer, short length)
{
	memset(buffer, 0x00, length);
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);

	// CDB
	m_cmd[0] = 0x12;					// operation code
	m_cmd[1] = 0x00;					// EVPD bit
	m_cmd[2] = 0x00;					// page or opcode (only if EVPD is set)
	m_cmd[3] = (length >> 8) & 0xFF;	// allocation length (MSB)
	m_cmd[4] = length & 0xFF;			// allocation length (LSB)

	// send to drive
	return (send_cmd(m_hHandle, m_cmd, buffer, 0, length) == 0);
}

/**
 * SEND DIAGNOSTIC command
 */
bool HDDrive::SendDiagnostic(uint8_t* buffer, short length)
{
	memset(m_cmd, 0x00, CDROM_PACKET_SIZE);
	m_cmd[0] = 0x1D;		// operation code (send diagnostic)
	m_cmd[4] = 0x08;		// parameter list length (LSB)

	// send to drive
	return (send_cmd(m_hHandle, m_cmd, buffer, length, 0) == 0);
}

/**
 *
 */
bool HDDrive::IsAuthenticated		()	{ return m_isAuthenticated; }
uint8_t* HDDrive::VolumeID			()	{ return m_VolumeID; }
