/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mediakeyblock.hpp"
#include "filesystem.hpp"
#include "rijndael.hpp"
#include "file.hpp"
#include <iostream>
#include <sstream>
#include <fstream>
#include <stdexcept>

MediaKeyBlock::MediaKeyBlock(Logger* pLog)
:	m_pLog(pLog),
	m_pucMKBFile(0),
	m_uiMKBSize(0),
	m_uiRecordLength(0),
	m_bIsProcessed(false),
	m_MediaKeyDataTable(0)
{
}

MediaKeyBlock::~MediaKeyBlock()
{
	if (m_pucMKBFile)			delete[] m_pucMKBFile;
	if (m_MediaKeyDataTable)	delete[] m_MediaKeyDataTable;
}

/**
 * Load Media Block file into memory
 *
 * Media Key Block Format
 * A Media Key Block is formatted as a sequence of contiguous Records. Each Record begins with a one-byte
 * Record Type field, followed by a three-byte Record Length field. The Record Type field value indicates the
 * type of the Record, and the Record Length field value indicates the number of bytes in the Record, including the
 * Record Type and the Record Length fields themselves. Record lengths are always multiples of 4 bytes.
 */
void MediaKeyBlock::Load(const std::string strMountPoint)
{
	m_pLog->LogInfo("Processing MKB");
	std::ifstream 	ifMKB;
	std::string 	strFileName = strMountPoint + "AACS" + SEPERATOR + "MKBROM.AACS";

	File MKB(strFileName);
	uint32_t uiMKBSize = MKB.FileSize();
	std::stringstream ssMsg;
	ssMsg << "MKB filesize:\t" << uiMKBSize;
	m_pLog->LogDebug(ssMsg.str());

	ifMKB.open(strFileName.c_str(), std::ios::binary | std::ios::in);
	if (!ifMKB.is_open()) {
		throw std::runtime_error("Error opening Media Key File " + strFileName);
	}

	try
	{
		m_pucMKBFile = new uint8_t[uiMKBSize];
	}
	catch (std::bad_alloc&) {
		throw std::runtime_error("Error loading MKB into memory. Could not allocate memory.");
	}

	ifMKB.read((char*)m_pucMKBFile, uiMKBSize);
/*
	if (!ifMKB.eof())
	{
		std::cerr << "Error reading Media Key File: could not read entire file." << std::endl;
		exit(0);
	}
*/
	ifMKB.close();

	//for (int i = 0; i < 12; i++)
	//	printf("0x%02X ", m_pucMKBFile[i]);

	// scan for records
	uint32_t	uiPos			= 0;
	eRecordType	eTypeFound 		= RECORD_UNKNOWN;
	m_uiRecordLength	 		= 0;

	while (uiPos < uiMKBSize)
	{
		eTypeFound 	 		 = static_cast<eRecordType>(m_pucMKBFile[uiPos]);
		m_uiRecordLength	 = m_pucMKBFile[uiPos + 3];
		m_uiRecordLength	+= m_pucMKBFile[uiPos + 2] * 256;
		m_uiRecordLength	+= m_pucMKBFile[uiPos + 1] * 256 * 256;
		//std::cout << "type=" << eTypeFound << ", uiRecordLength=" << m_uiRecordLength << std::endl;

		switch (eTypeFound)
		{
			case RECORD_VERSION:
				ExtractVersion(&m_pucMKBFile[uiPos]);
				break;
			case RECORD_HOST_RL:
				// ignore
				break;
			case RECORD_DRIVE_RL:
				// ignore
				break;
			case RECORD_VERIFY_MK:
				ExtractVerificationData(&m_pucMKBFile[uiPos]);
				break;
			case RECORD_EXPL_SUBSET:
				ExtractExplicitSubsetDifferences(&m_pucMKBFile[uiPos]);
				break;
			case RECORD_SUBSET_INDEX:
				// ignore
				break;
			case RECORD_MK_DATA:
				ExtractMediaKeyData(&m_pucMKBFile[uiPos], m_uiRecordLength);
				break;
			case RECORD_MKB_END:
				m_bIsProcessed = true;
				return;
		}

		// update pos
		uiPos += m_uiRecordLength;
	}

	m_bIsProcessed = true;
}

/**
 * Get MKB type and version
 *
 * Type 0x00031003 = normal MKB
 * Type 0x00041003 = MKB to use with Key Conversion Data
 */
void MediaKeyBlock::ExtractVersion(uint8_t* buffer)
{
	m_pLog->LogInfo("Extracting MKB version:");
	uint8_t* pu8Pos = buffer + 4;
	m_u32Type		= pu8Pos[0] << 24 | pu8Pos[1] << 16 | pu8Pos[2] << 8 | pu8Pos[3];
	m_u32Version	= pu8Pos[4] << 24 | pu8Pos[5] << 16 | pu8Pos[6] << 8 | pu8Pos[7];

	std::stringstream ssMsg;
	ssMsg << "\tType " << m_u32Type << ", Version\t" << m_u32Version;
	m_pLog->LogDebug(ssMsg.str());
}

/**
 * Get Encrypted Verification Data from MKB
 *
 * Bytes 4 through 19 of the Record contain the ciphertext value
 *	Dv = AES-128E (Km, 0123456789ABCDEF16 || XXXXXXXXXXXXXXXX16)
 * where XXXXXXXXXXXXXXXX16 is an arbitrary 8-byte value, and Km is the correct final Media Key value.
 */
void MediaKeyBlock::ExtractVerificationData(uint8_t* buffer)
{
	memcpy(m_VerificationData, buffer + 4, sizeof(m_VerificationData));
}

/**
 * Get encrypted C-values (Media Key Data Record) from MKB
 *
 * This record gives the associated encrypted Media Key Data for the subset-differences identified in the Explicit
 * Subset-Difference Record. Each subset-difference has its associated 16 bytes in this record, in the same order it
 * is encountered in the Explicit Subset-Difference Record. This 16-byte is the ciphertext value C in the Media Key
 * calculation.
 */
void MediaKeyBlock::ExtractMediaKeyData(uint8_t* buffer, const uint32_t uiLength)
{
	m_pLog->LogInfo("Extracting Media Key Data");

	// alloc mem
	try
	{
		m_MediaKeyDataTable = new uint8_t[uiLength - 4];
	}
	catch (std::bad_alloc &e) {
		throw std::runtime_error("Error allocating memory for Media Key Data table");
	}

	// copy into table
	memcpy(m_MediaKeyDataTable, buffer + 4, uiLength - 4);
}

/**
 * Get Explicit Subset-Difference records
 *
 * If a device encounters a u mask value whose high-order two bits are non-zero, without finding an applicable
 * subset, it may conclude it is revoked. In other words, if the u mask is not of the form b00xxxxxx, this marks the
 * end of the list.
 */
void MediaKeyBlock::ExtractExplicitSubsetDifferences(uint8_t* buffer)
{
	m_pLog->LogInfo("Extracting U masks and UV numbers");

	uint8_t*	pu8Pos		= buffer + 4;
	uint32_t	u32Length	= pu8Pos[0] << 16 | pu8Pos[1] << 8 | pu8Pos[2];
	int			iRecordNo	= 0;
	bool 		bEndReached	= false;

	while ((pu8Pos < buffer + u32Length) && (!bEndReached))
	{
		// 	revoked and end of list
		if ((pu8Pos[0] & 0xC0) != 0x00) {
			bEndReached = true;
		}

		uint8_t	 Umask 		= pu8Pos[0];
		uint32_t UVnumber	= 0x00000000;
		UVnumber			= pu8Pos[1] << 24 | pu8Pos[2] << 16 | pu8Pos[3] << 8 | pu8Pos[4];

		// add to map
		m_mUmasks.insert( std::make_pair(iRecordNo, Umask) );
		m_mUVnumbers.insert( std::make_pair(iRecordNo, UVnumber) );

		//printf("\t%02X:%08X\n", Umask, UVnumber);
		iRecordNo++;
		pu8Pos += 5;
	}
}

//============================================================================
// Accessor junk
//============================================================================
bool	MediaKeyBlock::IsProcessed()		{ return m_bIsProcessed; }
uint32_t MediaKeyBlock::Version()			{ return m_u32Version; }
uint32_t MediaKeyBlock::Type()				{ return m_u32Type; }
uint8_t* MediaKeyBlock::VerificationData()	{ return m_VerificationData; }
uint8_t* MediaKeyBlock::MediaKeyData(const int iRecordNo)		{ return m_MediaKeyDataTable + iRecordNo * 16; }
